"""Models """

from .edsmodel import EDSModel
from .edssemmodel import EDSSEMModel
from .edstemmodel import EDSTEMModel
from .eelsmodel import EELSModel

__all__ = [
    "EDSModel",
    "EDSSEMModel",
    "EDSTEMModel",
    "EELSModel",
]


def __dir__():
    return sorted(__all__)



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The exSpy developers
#
# This file is part of exSpy.
#
# exSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

import numbers
import logging

import numpy as np
import dask.array as da
import traits.api as t
from scipy import constants
from prettytable import PrettyTable
from matplotlib.collections import LineCollection

import hyperspy.api as hs
from hyperspy.signal import BaseSetMetadataItems, BaseSignal
from hyperspy._signals.signal1d import Signal1D, LazySignal1D
import hyperspy.axes
from hyperspy.components1d import PowerLaw
from hyperspy.misc.utils import display, isiterable, underline
from hyperspy.misc.math_tools import optimal_fft_size

from hyperspy.ui_registry import add_gui_method, DISPLAY_DT, TOOLKIT_DT
from hyperspy.utils.markers import Texts, Lines
from hyperspy.docstrings.signal1d import (
    CROP_PARAMETER_DOC,
    SPIKES_DIAGNOSIS_DOCSTRING,
    MASK_ZERO_LOSS_PEAK_WIDTH,
    SPIKES_REMOVAL_TOOL_DOCSTRING,
)
from hyperspy.docstrings.signal import (
    SHOW_PROGRESSBAR_ARG,
    NUM_WORKERS_ARG,
    SIGNAL_MASK_ARG,
    NAVIGATION_MASK_ARG,
    LAZYSIGNAL_DOC,
)

from exspy.docstrings.model import EELSMODEL_PARAMETERS
from exspy.misc.elements import elements as elements_db
from exspy.misc.eels.tools import get_edges_near_energy
from exspy.misc.eels.electron_inelastic_mean_free_path import (
    iMFP_Iakoubovskii,
    iMFP_angular_correction,
)
from exspy.signal_tools import EdgesRange


_logger = logging.getLogger(__name__)


@add_gui_method(toolkey="exspy.microscope_parameters_EELS")
class EELSTEMParametersUI(BaseSetMetadataItems):
    convergence_angle = t.Float(t.Undefined, label="Convergence semi-angle (mrad)")
    beam_energy = t.Float(t.Undefined, label="Beam energy (keV)")
    collection_angle = t.Float(t.Undefined, label="Collection semi-angle (mrad)")
    mapping = {
        "Acquisition_instrument.TEM.convergence_angle": "convergence_angle",
        "Acquisition_instrument.TEM.beam_energy": "beam_energy",
        "Acquisition_instrument.TEM.Detector.EELS.collection_angle": "collection_angle",
    }


class EELSSpectrum(Signal1D):

    """Signal class for EELS spectra."""

    _signal_type = "EELS"
    _alias_signal_types = ["TEM EELS"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Attributes defaults
        self.subshells = set()
        self.elements = set()
        self.edges = list()
        if hasattr(self.metadata, "Sample") and hasattr(
            self.metadata.Sample, "elements"
        ):
            self.add_elements(self.metadata.Sample.elements)
        self.axes_manager.signal_axes[0].is_binned = True
        self._edge_markers = {"names": [], "lines": None, "texts": None}

    def add_elements(self, elements, include_pre_edges=False):
        """Declare the elemental composition of the sample.

        The ionisation edges of the elements present in the current
        energy range will be added automatically.

        Parameters
        ----------
        elements : tuple of strings
            The symbol of the elements. Note this input must always be
            in the form of a tuple. Meaning: add_elements(('C',)) will
            work, while add_elements(('C')) will NOT work.
        include_pre_edges : bool
            If True, the ionization edges with an onset below the lower
            energy limit of the SI will be included

        Examples
        --------

        >>> s = hs.signals.EELSSpectrum(np.arange(1024))
        >>> s.add_elements(('C', 'O'))

        Raises
        ------
        ValueError

        """
        if not isiterable(elements) or isinstance(elements, str):
            raise ValueError(
                "Input must be in the form of a tuple. For example, "
                "if `s` is the variable containing this EELS spectrum:\n "
                ">>> s.add_elements(('C',))\n"
                "See the docstring for more information."
            )

        for element in elements:
            if isinstance(element, bytes):
                element = element.decode()
            if element in elements_db:
                self.elements.add(element)
            else:
                raise ValueError(
                    "%s is not a valid symbol of a chemical element" % element
                )
        if not hasattr(self.metadata, "Sample"):
            self.metadata.add_node("Sample")
        self.metadata.Sample.elements = list(self.elements)
        if self.elements:
            self.generate_subshells(include_pre_edges)

    def generate_subshells(self, include_pre_edges=False):
        """Calculate the subshells for the current energy range for the
        elements present in self.elements

        Parameters
        ----------
        include_pre_edges : bool
            If True, the ionization edges with an onset below the lower
            energy limit of the SI will be included

        """
        Eaxis = self.axes_manager.signal_axes[0].axis
        if not include_pre_edges:
            start_energy = Eaxis[0]
        else:
            start_energy = 0.0
        end_energy = Eaxis[-1]
        for element in self.elements:
            e_shells = list()
            for shell in elements_db[element]["Atomic_properties"]["Binding_energies"]:
                if shell[-1] != "a":
                    energy = elements_db[element]["Atomic_properties"][
                        "Binding_energies"
                    ][shell]["onset_energy (eV)"]
                    if start_energy <= energy <= end_energy:
                        subshell = "%s_%s" % (element, shell)
                        if subshell not in self.subshells:
                            self.subshells.add("%s_%s" % (element, shell))
                            e_shells.append(subshell)

    def edges_at_energy(
        self,
        energy="interactive",
        width=10,
        only_major=False,
        order="closest",
        display=True,
        toolkit=None,
    ):
        """Show EELS edges according to an energy range selected from the
        spectrum or within a provided energy window

        Parameters
        ----------
        energy : 'interactive' or float
            If it is 'interactive', a table with edges are shown and it depends
            on the energy range selected in the spectrum. If it is a float, a
            table with edges are shown and it depends on the energy window
            defined by energy +/- (width/2). The default is 'interactive'.
        width : float
            Width of window, in eV, around energy in which to find nearby
            energies, i.e. a value of 10 eV (the default) means to
            search +/- 5 eV. The default is 10.
        only_major : bool
            Whether to show only the major edges. The default is False.
        order : str
            Sort the edges, if 'closest', return in the order of energy
            difference, if 'ascending', return in ascending order, similarly
            for 'descending'. The default is 'closest'.

        Returns
        -------
        An interactive widget if energy is 'interactive', or a html-format
        table or ASCII table, depends on the environment.
        """

        if energy == "interactive":
            er = EdgesRange(self, interactive=True)
            return er.gui(display=display, toolkit=toolkit)
        else:
            self.print_edges_near_energy(energy, width, only_major, order)

    @staticmethod
    def print_edges_near_energy(
        energy=None, width=10, only_major=False, order="closest", edges=None
    ):
        """Find and print a table of edges near a given energy that are within
        the given energy window.

        Parameters
        ----------
        energy : float
            Energy to search, in eV
        width : float
            Width of window, in eV, around energy in which to find nearby
            energies, i.e. a value of 10 eV (the default) means to
            search +/- 5 eV. The default is 10.
        only_major : bool
            Whether to show only the major edges. The default is False.
        order : str
            Sort the edges, if 'closest', return in the order of energy
            difference, if 'ascending', return in ascending order, similarly
            for 'descending'. The default is 'closest'.
        edges : iterable
            A sequence of edges, if provided, it overrides energy, width,
            only_major and order.

        Returns
        -------
        A PrettyText object where its representation is ASCII in terminal and
        html-formatted in Jupyter notebook
        """

        if edges is None and energy is not None:
            edges = get_edges_near_energy(
                energy=energy, width=width, only_major=only_major, order=order
            )
        elif edges is None and energy is None:
            raise ValueError("Either energy or edges should be provided.")

        table = PrettyTable()
        table.field_names = ["edge", "onset energy (eV)", "relevance", "description"]

        for edge in edges:
            element, shell = edge.split("_")
            shell_dict = elements_db[element]["Atomic_properties"]["Binding_energies"][
                shell
            ]

            onset = shell_dict["onset_energy (eV)"]
            relevance = shell_dict["relevance"]
            threshold = shell_dict["threshold"]
            edge_ = shell_dict["edge"]
            description = threshold + ". " * (threshold != "" and edge_ != "") + edge_

            table.add_row([edge, onset, relevance, description])

        # this ensures the html version try its best to mimick the ASCII one
        table.format = True

        display(table)

    def estimate_zero_loss_peak_centre(self, mask=None):
        """Estimate the position of the zero-loss peak.

        This function provides just a coarse estimation of the position
        of the zero-loss peak centre by computing the position of the maximum
        of the spectra. For subpixel accuracy use `estimate_shift1D`.

        Parameters
        ----------
        mask : Signal1D of bool data type or bool array
            It must have signal_dimension = 0 and navigation_shape equal to the
            navigation shape of the current signal. Where mask is True the
            shift is not computed and set to nan.

        Returns
        -------
        zlpc : Signal1D subclass
            The estimated position of the maximum of the ZLP peak.

        Notes
        -----
        This function only works when the zero-loss peak is the most
        intense feature in the spectrum. If it is not in most cases
        the spectrum can be cropped to meet this criterion.
        Alternatively use `estimate_shift1D`.

        See Also
        --------
        estimate_shift1D, align_zero_loss_peak

        """
        self._check_signal_dimension_equals_one()
        self._check_navigation_mask(mask)
        if isinstance(mask, BaseSignal):
            mask = mask.data
        zlpc = self.valuemax(-1)
        if mask is not None:
            zlpc.data = np.where(mask, np.nan, zlpc.data)
        zlpc.set_signal_type("")
        title = self.metadata.General.title
        zlpc.metadata.General.title = "ZLP(%s)" % title
        return zlpc

    def align_zero_loss_peak(
        self,
        calibrate=True,
        also_align=[],
        print_stats=True,
        subpixel=True,
        mask=None,
        signal_range=None,
        show_progressbar=None,
        crop=True,
        **kwargs,
    ):
        """Align the zero-loss peak.

        This function first aligns the spectra using the result of
        `estimate_zero_loss_peak_centre` which finds the maximum in the
        given energy range, then if subpixel is True,
        proceeds to align with subpixel accuracy using `align1D`. The offset
        is automatically correct if `calibrate` is True.

        Parameters
        ----------
        calibrate : bool
            If True, set the offset of the spectral axis so that the
            zero-loss peak is at position zero.
        also_align : list of signals
            A list containing other spectra of identical dimensions to
            align using the shifts applied to the current spectrum.
            If `calibrate` is True, the calibration is also applied to
            the spectra in the list.
        print_stats : bool
            If True, print summary statistics of the ZLP maximum before
            the alignment.
        subpixel : bool
            If True, perform the alignment with subpixel accuracy
            using cross-correlation.
        mask : Signal1D of bool data type or bool array.
            It must have signal_dimension = 0 and navigation_shape equal to
            the shape of the current signal. Where mask is True the shift is
            not computed and set to nan.
        signal_range : tuple of integers, tuple of floats. Optional
            Will only search for the ZLP within the signal_range. If given
            in integers, the range will be in index values. If given floats,
            the range will be in spectrum values. Useful if there are features
            in the spectrum which are more intense than the ZLP.
            Default is searching in the whole signal. Note that ROIs can be used
            in place of a tuple.
        %s
        %s

        Raises
        ------
        NotImplementedError
            If the signal axis is a non-uniform axis.

        Examples
        --------
        >>> s_ll = hs.signals.EELSSpectrum(np.zeros(1000))
        >>> s_ll.data[100] = 100
        >>> s_ll.align_zero_loss_peak()

        Aligning both the lowloss signal and another signal

        >>> s = hs.signals.EELSSpectrum(np.range(1000))
        >>> s_ll.align_zero_loss_peak(also_align=[s])

        Aligning within a narrow range of the lowloss signal

        >>> s_ll.align_zero_loss_peak(signal_range=(-10.,10.))


        See Also
        --------
        estimate_zero_loss_peak_centre, align1D, estimate_shift1D.

        Notes
        -----
        Any extra keyword arguments are passed to `align1D`. For
        more information read its docstring.

        """

        def substract_from_offset(value, signals):
            # Test that axes is uniform
            if not self.axes_manager[-1].is_uniform:
                raise NotImplementedError(
                    "Support for EELS signals with "
                    "non-uniform signal axes is not yet implemented."
                )
            if isinstance(value, da.Array):
                value = value.compute()
            for signal in signals:
                signal.axes_manager[-1].offset -= value
                signal.events.data_changed.trigger(signal)

        def estimate_zero_loss_peak_centre(s, mask, signal_range):
            if signal_range:
                zlpc = s.isig[
                    signal_range[0] : signal_range[1]
                ].estimate_zero_loss_peak_centre(mask=mask)
            else:
                zlpc = s.estimate_zero_loss_peak_centre(mask=mask)
            return zlpc

        zlpc = estimate_zero_loss_peak_centre(
            self, mask=mask, signal_range=signal_range
        )

        mean_ = np.nanmean(zlpc.data)

        if print_stats is True:
            print(underline("Initial ZLP position statistics"))
            zlpc.print_summary_statistics()

        for signal in also_align + [self]:
            shift_array = -zlpc.data + mean_
            if zlpc._lazy:
                # We must compute right now because otherwise any changes to the
                # axes_manager of the signal later in the workflow may result in
                # a wrong shift_array
                shift_array = shift_array.compute()
            signal.shift1D(shift_array, crop=crop, show_progressbar=show_progressbar)

        if calibrate is True:
            zlpc = estimate_zero_loss_peak_centre(
                self, mask=mask, signal_range=signal_range
            )
            substract_from_offset(np.nanmean(zlpc.data), also_align + [self])

        if subpixel is False:
            return

        start, end = signal_range or (-3.0, 3.0)

        if calibrate is False:
            start += mean_
            end += mean_

        start = (
            start
            if start > self.axes_manager[-1].axis[0]
            else self.axes_manager[-1].axis[0]
        )
        end = (
            end
            if end < self.axes_manager[-1].axis[-1]
            else self.axes_manager[-1].axis[-1]
        )

        if self.axes_manager.navigation_size > 1:
            self.align1D(
                start,
                end,
                also_align=also_align,
                show_progressbar=show_progressbar,
                mask=mask,
                crop=crop,
                **kwargs,
            )
        if calibrate is True:
            zlpc = estimate_zero_loss_peak_centre(
                self, mask=mask, signal_range=signal_range
            )
            substract_from_offset(np.nanmean(zlpc.data), also_align + [self])

    align_zero_loss_peak.__doc__ %= (SHOW_PROGRESSBAR_ARG, CROP_PARAMETER_DOC)

    def get_zero_loss_peak_mask(self, zero_loss_peak_mask_width=5.0, signal_mask=None):
        """Return boolean array with True value at the position of the zero
        loss peak. This mask can be used to restrict operation to the signal
        locations not marked as True (masked).

        Parameters
        ----------
        zero_loss_peak_mask_width: float
            Width of the zero loss peak mask.
        %s

        Returns
        -------
        bool array
        """
        zlpc = self.estimate_zero_loss_peak_centre()
        (signal_axis,) = self.axes_manager[self.axes_manager.signal_axes]
        axis = signal_axis.axis
        mini_value = zlpc.data.mean() - zero_loss_peak_mask_width / 2
        maxi_value = zlpc.data.mean() + zero_loss_peak_mask_width / 2
        mask = np.logical_and(mini_value <= axis, axis <= maxi_value)
        if signal_mask is not None:
            signal_mask = np.logical_or(mask, signal_mask)
        else:
            signal_mask = mask
        return signal_mask

    get_zero_loss_peak_mask.__doc__ %= SIGNAL_MASK_ARG

    def spikes_diagnosis(
        self,
        signal_mask=None,
        navigation_mask=None,
        zero_loss_peak_mask_width=None,
        **kwargs,
    ):
        if zero_loss_peak_mask_width is not None:
            signal_mask = self.get_zero_loss_peak_mask(
                zero_loss_peak_mask_width, signal_mask
            )
        super().spikes_diagnosis(
            signal_mask=signal_mask, navigation_mask=None, **kwargs
        )

    spikes_diagnosis.__doc__ = SPIKES_DIAGNOSIS_DOCSTRING % MASK_ZERO_LOSS_PEAK_WIDTH

    def spikes_removal_tool(
        self,
        signal_mask=None,
        navigation_mask=None,
        threshold="auto",
        zero_loss_peak_mask_width=None,
        interactive=True,
        display=True,
        toolkit=None,
    ):
        if zero_loss_peak_mask_width is not None:
            axis = self.axes_manager.signal_axes[0].axis
            # check the zero_loss is in the signal
            if (
                axis[0] - zero_loss_peak_mask_width / 2 > 0
                or axis[-1] + zero_loss_peak_mask_width / 2 < 0
            ):
                raise ValueError("The zero loss peaks isn't in the energy range.")
            signal_mask = self.get_zero_loss_peak_mask(
                zero_loss_peak_mask_width, signal_mask
            )
        super().spikes_removal_tool(
            signal_mask=signal_mask,
            navigation_mask=navigation_mask,
            threshold=threshold,
            interactive=interactive,
            display=display,
            toolkit=toolkit,
        )

    spikes_removal_tool.__doc__ = SPIKES_REMOVAL_TOOL_DOCSTRING % (
        SIGNAL_MASK_ARG,
        NAVIGATION_MASK_ARG,
        MASK_ZERO_LOSS_PEAK_WIDTH,
        DISPLAY_DT,
        TOOLKIT_DT,
    )

    def estimate_elastic_scattering_intensity(self, threshold, show_progressbar=None):
        """Rough estimation of the elastic scattering intensity by
        truncation of a EELS low-loss spectrum.

        Parameters
        ----------
        threshold : {Signal1D, float, int}
            Truncation energy to estimate the intensity of the elastic
            scattering. The threshold can be provided as a signal of the same
            dimension as the input spectrum navigation space containing the
            threshold value in the energy units. Alternatively a constant
            threshold can be specified in energy/index units by passing
            float/int.
        %s

        Returns
        -------
        I0: Signal1D
            The elastic scattering intensity.

        See Also
        --------
        estimate_elastic_scattering_threshold

        """
        # TODO: Write units tests
        self._check_signal_dimension_equals_one()

        if show_progressbar is None:
            show_progressbar = hs.preferences.General.show_progressbar

        if isinstance(threshold, numbers.Number):
            I0 = self.isig[:threshold].integrate1D(-1)
        else:
            ax = self.axes_manager.signal_axes[0]
            # I0 = self._get_navigation_signal()
            # I0 = I0.transpose(signal_axes=[])
            threshold = threshold.transpose(signal_axes=[])
            binned = ax.is_binned

            def estimating_function(data, threshold=None):
                if np.isnan(threshold):
                    return np.nan
                else:
                    # the object is just an array, so have to reimplement
                    # integrate1D. However can make certain assumptions, for
                    # example 1D signal and pretty much always binned. Should
                    # probably at some point be joint
                    ind = ax.value2index(threshold)
                    data = data[:ind]
                    if binned:
                        return data.sum()
                    else:
                        from scipy.integrate import simps

                        axis = ax.axis[:ind]
                        return simps(y=data, x=axis)

            I0 = self.map(
                estimating_function,
                threshold=threshold,
                ragged=False,
                show_progressbar=show_progressbar,
                inplace=False,
            )
        I0.metadata.General.title = self.metadata.General.title + " elastic intensity"
        I0.set_signal_type("")
        if self.tmp_parameters.has_item("filename"):
            I0.tmp_parameters.filename = (
                self.tmp_parameters.filename + "_elastic_intensity"
            )
            I0.tmp_parameters.folder = self.tmp_parameters.folder
            I0.tmp_parameters.extension = self.tmp_parameters.extension
        return I0

    estimate_elastic_scattering_intensity.__doc__ %= SHOW_PROGRESSBAR_ARG

    def estimate_elastic_scattering_threshold(
        self, window=10.0, tol=None, window_length=5, polynomial_order=3, start=1.0
    ):
        """Calculate the first inflexion point of the spectrum derivative
        within a window.

        This method assumes that the zero-loss peak is located at position zero
        in all the spectra. Currently it looks for an inflexion point, that can
        be a local maximum or minimum. Therefore, to estimate the elastic
        scattering threshold `start` + `window` must be less than the first
        maximum for all spectra (often the bulk plasmon maximum). If there is
        more than one inflexion point in energy the window it selects the
        smoother one what, often, but not always, is a good choice in this
        case.

        Parameters
        ----------
        window : {None, float}
            If None, the search for the local inflexion point is performed
            using the full energy range. A positive float will restrict
            the search to the (0,window] energy window, where window is given
            in the axis units. If no inflexion point is found in this
            spectral range the window value is returned instead.
        tol : {None, float}
            The threshold tolerance for the derivative. If "auto" it is
            automatically calculated as the minimum value that guarantees
            finding an inflexion point in all the spectra in given energy
            range.
        window_length : int
            If non zero performs order three Savitzky-Golay smoothing
            to the data to avoid falling in local minima caused by
            the noise. It must be an odd integer.
        polynomial_order : int
            Savitzky-Golay filter polynomial order.
        start : float
            Position from the zero-loss peak centre from where to start
            looking for the inflexion point.


        Returns
        -------

        threshold : Signal1D
            A Signal1D of the same dimension as the input spectrum
            navigation space containing the estimated threshold. Where the
            threshold couldn't be estimated the value is set to nan.

        See Also
        --------

        estimate_elastic_scattering_intensity,align_zero_loss_peak,
        find_peaks1D_ohaver, fourier_ratio_deconvolution.

        Notes
        -----

        The main purpose of this method is to be used as input for
        `estimate_elastic_scattering_intensity`. Indeed, for currently
        achievable energy resolutions, there is not such a thing as a elastic
        scattering threshold. Therefore, please be aware of the limitations of
        this method when using it.

        """
        self._check_signal_dimension_equals_one()
        # Create threshold with the same shape as the navigation dims.
        threshold = self._get_navigation_signal().transpose(signal_axes=0)

        # Progress Bar
        axis = self.axes_manager.signal_axes[0]
        min_index, max_index = axis.value_range_to_indices(start, start + window)
        if max_index < min_index + 10:
            raise ValueError("Please select a bigger window")
        s = self.isig[min_index:max_index].deepcopy()
        if window_length:
            s.smooth_savitzky_golay(
                polynomial_order=polynomial_order,
                window_length=window_length,
                differential_order=1,
            )
        else:
            s = s.derivative(-1)
        if tol is None:
            tol = np.max(abs(s.data).min(axis.index_in_array))
        saxis = s.axes_manager[-1]
        inflexion = (abs(s.data) <= tol).argmax(saxis.index_in_array)
        if isinstance(inflexion, da.Array):
            inflexion = inflexion.compute()
        threshold.data[:] = saxis.index2value(inflexion)
        if isinstance(inflexion, np.ndarray):
            threshold.data[inflexion == 0] = np.nan
        else:  # Single spectrum
            if inflexion == 0:
                threshold.data[:] = np.nan
        del s
        if np.isnan(threshold.data).any():
            _logger.warning(
                "No inflexion point could be found in some positions "
                "that have been marked with nans."
            )
        # Create spectrum image, stop and return value
        threshold.metadata.General.title = (
            self.metadata.General.title + " elastic scattering threshold"
        )
        if self.tmp_parameters.has_item("filename"):
            threshold.tmp_parameters.filename = (
                self.tmp_parameters.filename + "_elastic_scattering_threshold"
            )
            threshold.tmp_parameters.folder = self.tmp_parameters.folder
            threshold.tmp_parameters.extension = self.tmp_parameters.extension
        threshold.set_signal_type("")
        return threshold

    def estimate_thickness(
        self,
        threshold=None,
        zlp=None,
        density=None,
        mean_free_path=None,
    ):
        """Estimates the thickness (relative and absolute)
        of a sample using the log-ratio method.

        The current EELS spectrum must be a low-loss spectrum containing
        the zero-loss peak. The hyperspectrum must be well calibrated
        and aligned. To obtain the thickness relative to the mean free path
        don't set the `density` and the `mean_free_path`.

        Parameters
        ----------
        threshold : {BaseSignal, float}, optional
            If the zero-loss-peak is not provided, use this energy threshold
            to roughly estimate its intensity by truncation.
            If the threshold is constant across the dataset use a float. Otherwise,
            provide a signal of
            the same dimension as the input spectrum navigation space
            containing the threshold value in the energy units.
        zlp : BaseSignal, optional
            If not None the zero-loss peak intensity is calculated from the ZLP
            spectrum supplied by integration.
        mean_free_path : float, optional
            The mean free path of the material in nanometers.
            If not provided, the thickness
            is given relative to the mean free path.
        density : float, optional
            The density of the material in g/cm**3. This is used to estimate the mean
            free path when the mean free path is not known and to perform the
            angular corrections.

        Returns
        -------
        s : BaseSignal
            The thickness relative to the MFP. It returns a Signal1D,
            Signal2D or a BaseSignal, depending on the current navigation
            dimensions.

        Notes
        -----
        For details see Egerton, R. Electron Energy-Loss Spectroscopy in the Electron
        Microscope.  Springer-Verlag, 2011.
        """
        axis = self.axes_manager.signal_axes[0]
        total_intensity = self.integrate1D(axis.index_in_array).data
        if threshold is None and zlp is None:
            raise ValueError(
                "Please provide one of the following keywords: " "`threshold`, `zlp`"
            )
        if zlp is not None:
            I0 = zlp.integrate1D(axis.index_in_array).data
        else:
            I0 = self.estimate_elastic_scattering_intensity(
                threshold=threshold,
            ).data

        t_over_lambda = np.log(total_intensity / I0)

        if density is not None:
            if self._are_microscope_parameters_missing():
                raise RuntimeError(
                    "Some microscope parameters are missing. Please use the "
                    "`set_microscope_parameters()` method to set them. "
                    "If you don't know them, don't set the `density` keyword."
                )
            else:
                md = self.metadata.Acquisition_instrument.TEM
                t_over_lambda *= iMFP_angular_correction(
                    beam_energy=md.beam_energy,
                    alpha=md.convergence_angle,
                    beta=md.Detector.EELS.collection_angle,
                    density=density,
                )
                if mean_free_path is None:
                    mean_free_path = iMFP_Iakoubovskii(
                        electron_energy=self.metadata.Acquisition_instrument.TEM.beam_energy,
                        density=density,
                    )
                    _logger.info(f"The estimated iMFP is {mean_free_path} nm")
        else:
            _logger.warning(
                "Computing the thickness without taking into account the effect of "
                "the limited collection angle, what usually leads to underestimating "
                "the thickness. To perform the angular corrections you must provide "
                "the density of the material."
            )

        s = self._get_navigation_signal(data=t_over_lambda)
        if mean_free_path is not None:
            s.data *= mean_free_path
            s.metadata.General.title = self.metadata.General.title + " thickness (nm)"
            s.metadata.Signal.quantity = "thickness (nm)"
        else:
            _logger.warning(
                "Computing the relative thickness. To compute the absolute "
                "thickness provide the `mean_free_path` and/or the `density`"
            )
            s.metadata.General.title = (
                self.metadata.General.title + " $\\frac{t}{\\lambda}$"
            )
            s.metadata.Signal.quantity = "$\\frac{t}{\\lambda}$"
        if self.tmp_parameters.has_item("filename"):
            s.tmp_parameters.filename = self.tmp_parameters.filename + "_thickness"
            s.tmp_parameters.folder = self.tmp_parameters.folder
            s.tmp_parameters.extension = self.tmp_parameters.extension
        s = s.transpose(signal_axes=[])
        s.set_signal_type("")
        return s

    def fourier_log_deconvolution(self, zlp, add_zlp=False, crop=False):
        """Performs fourier-log deconvolution.

        Parameters
        ----------
        zlp : EELSSpectrum
            The corresponding zero-loss peak.

        add_zlp : bool
            If True, adds the ZLP to the deconvolved spectrum
        crop : bool
            If True crop the spectrum to leave out the channels that
            have been modified to decay smoothly to zero at the sides
            of the spectrum.

        Returns
        -------
        An EELSSpectrum containing the current data deconvolved.

        Raises
        ------
        NotImplementedError
            If the signal axis is a non-uniform axis.

        Notes
        -----
        For details see: Egerton, R. Electron Energy-Loss
        Spectroscopy in the Electron Microscope. Springer-Verlag, 2011.

        """
        self._check_signal_dimension_equals_one()
        if not self.axes_manager.signal_axes[0].is_uniform:
            raise NotImplementedError(
                "This operation is not yet implemented for non-uniform energy axes"
            )
        s = self.deepcopy()
        zlp_size = zlp.axes_manager.signal_axes[0].size
        self_size = self.axes_manager.signal_axes[0].size
        tapped_channels = s.hanning_taper()
        # Conservative new size to solve the wrap-around problem
        size = zlp_size + self_size - 1
        # Calculate optimal FFT padding for performance
        complex_result = zlp.data.dtype.kind == "c" or s.data.dtype.kind == "c"
        size = optimal_fft_size(size, not complex_result)

        axis = self.axes_manager.signal_axes[0]

        z = np.fft.rfft(zlp.data, n=size, axis=axis.index_in_array)
        j = np.fft.rfft(s.data, n=size, axis=axis.index_in_array)
        if self._lazy or zlp._lazy:
            j1 = z * da.log(j / z).map_blocks(np.nan_to_num)
        else:
            j1 = z * np.nan_to_num(np.log(j / z))
        sdata = np.fft.irfft(j1, axis=axis.index_in_array)

        s.data = sdata[
            s.axes_manager._get_data_slice(
                [
                    (axis.index_in_array, slice(None, self_size)),
                ]
            )
        ]
        if add_zlp is True:
            if self_size >= zlp_size:
                if self._lazy:
                    _slices_before = s.axes_manager._get_data_slice(
                        [
                            (axis.index_in_array, slice(None, zlp_size)),
                        ]
                    )
                    _slices_after = s.axes_manager._get_data_slice(
                        [
                            (axis.index_in_array, slice(zlp_size, None)),
                        ]
                    )
                    s.data = da.stack(
                        (s.data[_slices_before] + zlp.data, s.data[_slices_after]),
                        axis=axis.index_in_array,
                    )
                else:
                    s.data[
                        s.axes_manager._get_data_slice(
                            [
                                (axis.index_in_array, slice(None, zlp_size)),
                            ]
                        )
                    ] += zlp.data
            else:
                s.data += zlp.data[
                    s.axes_manager._get_data_slice(
                        [
                            (axis.index_in_array, slice(None, self_size)),
                        ]
                    )
                ]

        s.metadata.General.title = (
            s.metadata.General.title + " after Fourier-log deconvolution"
        )
        if s.tmp_parameters.has_item("filename"):
            s.tmp_parameters.filename = (
                self.tmp_parameters.filename + "_after_fourier_log_deconvolution"
            )
        if crop is True:
            s.crop(axis.index_in_axes_manager, None, int(-tapped_channels))
        return s

    def fourier_ratio_deconvolution(
        self,
        ll,
        fwhm=None,
        threshold=None,
        extrapolate_lowloss=True,
        extrapolate_coreloss=True,
    ):
        """Performs Fourier-ratio deconvolution.

        The core-loss should have the background removed. To reduce the noise
        amplification the result is convolved with a Gaussian function.

        Parameters
        ----------
        ll: EELSSpectrum
            The corresponding low-loss (ll) EELSSpectrum.
        fwhm : float or None
            Full-width half-maximum of the Gaussian function by which
            the result of the deconvolution is convolved. It can be
            used to select the final SNR and spectral resolution. If
            None, the FWHM of the zero-loss peak of the low-loss is
            estimated and used.
        threshold : {None, float}
            Truncation energy to estimate the intensity of the
            elastic scattering. If None the threshold is taken as the
            first minimum after the ZLP centre.
        extrapolate_lowloss, extrapolate_coreloss : bool
            If True the signals are extrapolated using a power law,

        Raises
        ------
        NotImplementedError
            If the signal axis is a non-uniform axis.

        Notes
        -----
        For details see: Egerton, R. Electron Energy-Loss
        Spectroscopy in the Electron Microscope. Springer-Verlag, 2011.

        """
        self._check_signal_dimension_equals_one()
        if not self.axes_manager.signal_axes[0].is_uniform:
            raise NotImplementedError(
                "This operation is not yet implemented for non-uniform energy axes."
            )
        if not ll.axes_manager.signal_axes[0].is_uniform:
            raise NotImplementedError(
                "The low-loss energy axis is non-uniform. "
                "This operation is not yet implemented for non-uniform energy axes"
            )
        orig_cl_size = self.axes_manager.signal_axes[0].size

        if threshold is None:
            threshold = ll.estimate_elastic_scattering_threshold()

        if extrapolate_coreloss is True:
            cl = self.power_law_extrapolation(window_size=20, extrapolation_size=100)
        else:
            cl = self.deepcopy()

        if extrapolate_lowloss is True:
            ll = ll.power_law_extrapolation(window_size=100, extrapolation_size=100)
        else:
            ll = ll.deepcopy()

        ll.hanning_taper()
        cl.hanning_taper()

        ll_size = ll.axes_manager.signal_axes[0].size
        cl_size = self.axes_manager.signal_axes[0].size
        # Conservative new size to solve the wrap-around problem
        size = ll_size + cl_size - 1
        # Calculate the optimal FFT size
        size = optimal_fft_size(size)

        axis = ll.axes_manager.signal_axes[0]
        if fwhm is None:
            fwhm = float(
                ll.get_current_signal().estimate_peak_width()._get_current_data()
            )
            _logger.info("FWHM = %1.2f" % fwhm)

        I0 = ll.estimate_elastic_scattering_intensity(threshold=threshold)
        I0 = I0.data
        if ll.axes_manager.navigation_size > 0:
            I0_shape = list(I0.shape)
            I0_shape.insert(axis.index_in_array, 1)
            I0 = I0.reshape(I0_shape)

        from hyperspy.components1d import Gaussian

        g = Gaussian()
        g.sigma.value = fwhm / 2.3548
        g.A.value = 1
        g.centre.value = 0
        zl = g.function(
            np.linspace(axis.offset, axis.offset + axis.scale * (size - 1), size)
        )
        z = np.fft.rfft(zl)
        jk = np.fft.rfft(cl.data, n=size, axis=axis.index_in_array)
        jl = np.fft.rfft(ll.data, n=size, axis=axis.index_in_array)
        zshape = [
            1,
        ] * len(cl.data.shape)
        zshape[axis.index_in_array] = jk.shape[axis.index_in_array]
        cl.data = np.fft.irfft(z.reshape(zshape) * jk / jl, axis=axis.index_in_array)
        cl.data *= I0
        cl.crop(-1, None, int(orig_cl_size))
        cl.metadata.General.title = (
            self.metadata.General.title + " after Fourier-ratio deconvolution"
        )
        if cl.tmp_parameters.has_item("filename"):
            cl.tmp_parameters.filename = (
                self.tmp_parameters.filename + "after_fourier_ratio_deconvolution"
            )
        return cl

    def richardson_lucy_deconvolution(
        self, psf, iterations=15, show_progressbar=None, num_workers=None
    ):
        """1D Richardson-Lucy Poissonian deconvolution of
        the spectrum by the given kernel.

        Parameters
        ----------
        psf : EELSSpectrum
            It must have the same signal dimension as the current
            spectrum and a spatial dimension of 0 or the same as the
            current spectrum.
        iterations : int
            Number of iterations of the deconvolution. Note that
            increasing the value will increase the noise amplification.
        %s
        %s

        Raises
        ------
        NotImplementedError
            If the signal axis is a non-uniform axis.

        Notes
        -----
        For details on the algorithm see Gloter, A., A. Douiri,
        M. Tence, and C. Colliex. “Improving Energy Resolution of
        EELS Spectra: An Alternative to the Monochromator Solution.”
        Ultramicroscopy 96, no. 3–4 (September 2003): 385–400.

        """
        if not self.axes_manager.signal_axes[0].is_uniform:
            raise NotImplementedError(
                "This operation is not yet implemented for non-uniform energy axes."
            )
        if show_progressbar is None:
            show_progressbar = hs.preferences.General.show_progressbar
        self._check_signal_dimension_equals_one()
        psf_size = psf.axes_manager.signal_axes[0].size
        maxval = self.axes_manager.navigation_size
        show_progressbar = show_progressbar and (maxval > 0)

        def deconv_function(signal, kernel=None, iterations=15, psf_size=None):
            imax = kernel.argmax()
            result = np.array(signal).copy()
            mimax = psf_size - 1 - imax
            for _ in range(iterations):
                first = np.convolve(kernel, result)[imax : imax + psf_size]
                result *= np.convolve(kernel[::-1], signal / first)[
                    mimax : mimax + psf_size
                ]
            return result

        ds = self.map(
            deconv_function,
            kernel=psf,
            iterations=iterations,
            psf_size=psf_size,
            show_progressbar=show_progressbar,
            num_workers=num_workers,
            ragged=False,
            inplace=False,
        )

        ds.metadata.General.title += (
            " after Richardson-Lucy deconvolution %i iterations" % iterations
        )
        if ds.tmp_parameters.has_item("filename"):
            ds.tmp_parameters.filename += "_after_R-L_deconvolution_%iiter" % iterations
        return ds

    richardson_lucy_deconvolution.__doc__ %= (SHOW_PROGRESSBAR_ARG, NUM_WORKERS_ARG)

    def _are_microscope_parameters_missing(self, ignore_parameters=[]):
        """
        Check if the EELS parameters necessary to calculate the GOS
        are defined in metadata. If not, in interactive mode
        raises an UI item to fill the values.
        The `ignore_parameters` list can be to ignore parameters.
        """
        must_exist = (
            "Acquisition_instrument.TEM.convergence_angle",
            "Acquisition_instrument.TEM.beam_energy",
            "Acquisition_instrument.TEM.Detector.EELS.collection_angle",
        )
        missing_parameters = []
        for item in must_exist:
            exists = self.metadata.has_item(item)
            if exists is False and item.split(".")[-1] not in ignore_parameters:
                missing_parameters.append(item)
        if missing_parameters:
            _logger.info("Missing parameters {}".format(missing_parameters))
            return True
        else:
            return False

    def set_microscope_parameters(
        self,
        beam_energy=None,
        convergence_angle=None,
        collection_angle=None,
        toolkit=None,
        display=True,
    ):
        if set((beam_energy, convergence_angle, collection_angle)) == {None}:
            tem_par = EELSTEMParametersUI(self)
            return tem_par.gui(toolkit=toolkit, display=display)
        mp = self.metadata
        if beam_energy is not None:
            mp.set_item("Acquisition_instrument.TEM.beam_energy", beam_energy)
        if convergence_angle is not None:
            mp.set_item(
                "Acquisition_instrument.TEM.convergence_angle", convergence_angle
            )
        if collection_angle is not None:
            mp.set_item(
                "Acquisition_instrument.TEM.Detector.EELS.collection_angle",
                collection_angle,
            )

    set_microscope_parameters.__doc__ = """
        Set the microscope parameters that are necessary to calculate
        the GOS.

        If not all of them are defined, in interactive mode
        raises an UI item to fill the values.

        beam_energy: float
            The energy of the electron beam in keV.
        convergence_angle : float
            The microscope convergence semi-angle in mrad.
        collection_angle : float
            The collection semi-angle in mrad.
        {}
        {}
        """.format(
        TOOLKIT_DT, DISPLAY_DT
    )

    def power_law_extrapolation(
        self, window_size=20, extrapolation_size=1024, add_noise=False, fix_neg_r=False
    ):
        """
        Extrapolate the spectrum to the right using a powerlaw.

        Parameters
        ----------
        window_size : int
            The number of channels from the right side of the
            spectrum that are used to estimate the power law
            parameters.
        extrapolation_size : int
            Size of the extrapolation in number of channels
        add_noise : bool
            If True, add poissonian noise to the extrapolated spectrum.
        fix_neg_r : bool
            If True, the negative values for the "components.PowerLaw"
            parameter r will be flagged and the extrapolation will be
            done with a constant zero-value.

        Returns
        -------
        A new spectrum, with the extrapolation.

        """
        self._check_signal_dimension_equals_one()
        axis = self.axes_manager.signal_axes[0]
        s = self.deepcopy()
        s.metadata.General.title += " %i channels extrapolated" % extrapolation_size
        if s.tmp_parameters.has_item("filename"):
            s.tmp_parameters.filename += (
                "_%i_channels_extrapolated" % extrapolation_size
            )
        new_shape = list(self.data.shape)
        new_shape[axis.index_in_array] += extrapolation_size
        if self._lazy:
            left_data = s.data
            right_shape = list(self.data.shape)
            right_shape[axis.index_in_array] = extrapolation_size
            right_chunks = list(self.data.chunks)
            right_chunks[axis.index_in_array] = (extrapolation_size,)
            right_data = da.zeros(
                shape=tuple(right_shape),
                chunks=tuple(right_chunks),
                dtype=self.data.dtype,
            )
            s.data = da.concatenate([left_data, right_data], axis=axis.index_in_array)
        else:
            # just old code
            s.data = np.zeros(new_shape)
            s.data[..., : axis.size] = self.data
        s.get_dimensions_from_data()
        pl = PowerLaw()
        pl._axes_manager = self.axes_manager
        A, r = pl.estimate_parameters(
            s,
            axis.index2value(axis.size - window_size),
            axis.index2value(axis.size - 1),
            out=True,
        )
        if fix_neg_r is True:
            A = np.where(r <= 0, 0, A)
        # If the signal is binned we need to bin the extrapolated power law
        # what, in a first approximation, can be done by multiplying by the
        # axis step size.
        if self.axes_manager[-1].is_binned:
            factor = s.axes_manager[-1].scale
        else:
            factor = 1
        if self._lazy:
            # only need new axes if the navigation dimension is not 0
            if s.axes_manager.navigation_dimension:
                rightslice = (..., None)
                axisslice = (None, slice(axis.size, None))
            else:
                rightslice = (...,)
                axisslice = (slice(axis.size, None),)
            right_chunks[axis.index_in_array] = 1
            x = da.from_array(
                s.axes_manager.signal_axes[0].axis[axisslice],
                chunks=(extrapolation_size,),
            )
            A = A[rightslice]
            r = r[rightslice]
            right_data = factor * A * x ** (-r)
            s.data = da.concatenate([left_data, right_data], axis=axis.index_in_array)
        else:
            s.data[..., axis.size :] = (
                factor
                * A[..., np.newaxis]
                * s.axes_manager.signal_axes[0].axis[np.newaxis, axis.size :]
                ** (-r[..., np.newaxis])
            )
        return s

    def kramers_kronig_analysis(
        self, zlp=None, iterations=1, n=None, t=None, delta=0.5, full_output=False
    ):
        r"""
        Calculate the complex dielectric function from a single scattering
        distribution (SSD) using the Kramers-Kronig relations.

        It uses the FFT method as in [1]_.  The SSD is an
        EELSSpectrum instance containing SSD low-loss EELS with no zero-loss
        peak. The internal loop is devised to approximately subtract the
        surface plasmon contribution supposing an unoxidized planar surface and
        neglecting coupling between the surfaces. This method does not account
        for retardation effects, instrumental broadening and surface plasmon
        excitation in particles.

        Note that either refractive index or thickness are required.
        If both are None or if both are provided an exception is raised.

        Parameters
        ----------
        zlp : {None, number, Signal1D}
            ZLP intensity. It is optional (can be None) if `t` is None and `n`
            is not None and the thickness estimation is not required. If `t`
            is not None, the ZLP is required to perform the normalization and
            if `t` is not None, the ZLP is required to calculate the thickness.
            If the ZLP is the same for all spectra, the integral of the ZLP
            can be provided as a number. Otherwise, if the ZLP intensity is not
            the same for all spectra, it can be provided as i) a Signal1D
            of the same dimensions as the current signal containing the ZLP
            spectra for each location ii) a BaseSignal of signal dimension 0
            and navigation_dimension equal to the current signal containing the
            integrated ZLP intensity.
        iterations : int
            Number of the iterations for the internal loop to remove the
            surface plasmon contribution. If 1 the surface plasmon contribution
            is not estimated and subtracted (the default is 1).
        n : {None, float}
            The medium refractive index. Used for normalization of the
            SSD to obtain the energy loss function. If given the thickness
            is estimated and returned. It is only required when `t` is None.
        t : {None, number, Signal1D}
            The sample thickness in nm. Used for normalization of the SSD
            to obtain the energy loss function. It is only required when
            `n` is None. If the thickness is the same for all spectra it can be
            given by a number. Otherwise, it can be provided as a BaseSignal
            with signal dimension 0 and navigation_dimension equal to the
            current signal.
        delta : float
            A small number (0.1-0.5 eV) added to the energy axis in
            specific steps of the calculation the surface loss correction to
            improve stability.
        full_output : bool
            If True, return a dictionary that contains the estimated
            thickness if `t` is None and the estimated surface plasmon
            excitation and the spectrum corrected from surface plasmon
            excitations if `iterations` > 1.

        Returns
        -------
        eps: DielectricFunction instance
            The complex dielectric function results,

                .. math::
                    \epsilon = \epsilon_1 + i*\epsilon_2,

            contained in an DielectricFunction instance.
        output: Dictionary (optional)
            A dictionary of optional outputs with the following keys

            * ``thickness``: the estimated  thickness in nm calculated by
              normalization of the SSD (only when ``t`` is None)
            * ``surface plasmon estimation``: the estimated surface plasmon
              excitation (only if ``iterations`` > 1.)

        Raises
        ------
        ValueError
            If both `n` and `t` are undefined (None).
        AttributeError
            If the beam_energy or the collection semi-angle are not defined in
            metadata.
        NotImplementedError
            If the signal axis is a non-uniform axis.

        Notes
        -----
        This method is based in Egerton's Matlab code [1]_ with a
        minor difference: the wrap-around problem when computing the FFTs is
        workarounded by padding the signal instead of subtracting the
        reflected tail.

        .. [1] Ray Egerton, "Electron Energy-Loss Spectroscopy in the Electron
           Microscope", Springer-Verlag, 2011.

        """
        if not self.axes_manager.signal_axes[0].is_uniform:
            raise NotImplementedError(
                "This operation is not yet implemented for non-uniform energy axes."
            )
        output = {}
        if iterations == 1:
            # In this case s.data is not modified so there is no need to make
            # a deep copy.
            s = self.isig[0.0:]
        else:
            s = self.isig[0.0:].deepcopy()

        sorig = self.isig[0.0:]
        # Avoid singularity at 0
        if s.axes_manager.signal_axes[0].axis[0] == 0:
            s = s.isig[1:]
            sorig = self.isig[1:]

        # Constants and units
        me = constants.value("electron mass energy equivalent in MeV") * 1e3  # keV

        # Mapped parameters
        self._are_microscope_parameters_missing(ignore_parameters=["convergence_angle"])
        e0 = s.metadata.Acquisition_instrument.TEM.beam_energy
        beta = s.metadata.Acquisition_instrument.TEM.Detector.EELS.collection_angle

        axis = s.axes_manager.signal_axes[0]
        eaxis = axis.axis.copy()

        if isinstance(zlp, hyperspy.signal.BaseSignal):
            if (
                zlp.axes_manager.navigation_dimension
                == self.axes_manager.navigation_dimension
            ):
                if zlp.axes_manager.signal_dimension == 0:
                    i0 = zlp.data
                else:
                    i0 = zlp.integrate1D(axis.index_in_axes_manager).data
            else:
                raise ValueError(
                    "The ZLP signal dimensions are not "
                    "compatible with the dimensions of the "
                    "low-loss signal"
                )
            # The following prevents errors if the signal is a single spectrum
            if len(i0) != 1:
                i0 = i0.reshape(np.insert(i0.shape, axis.index_in_array, 1))
        elif isinstance(zlp, numbers.Number):
            i0 = zlp
        else:
            raise ValueError(
                "The zero-loss peak input is not valid, it must be\
                             in the BaseSignal class or a Number."
            )

        if isinstance(t, hyperspy.signal.BaseSignal):
            if (
                t.axes_manager.navigation_dimension
                == self.axes_manager.navigation_dimension
            ) and (t.axes_manager.signal_dimension == 0):
                t = t.data
                t = t.reshape(np.insert(t.shape, axis.index_in_array, 1))
            else:
                raise ValueError(
                    "The thickness signal dimensions are not "
                    "compatible with the dimensions of the "
                    "low-loss signal"
                )
        elif isinstance(t, np.ndarray) and t.shape and t.shape != (1,):
            raise ValueError(
                "thickness must be a HyperSpy signal or a number," " not a NumPy array."
            )

        # Slicer to get the signal data from 0 to axis.size
        slicer = s.axes_manager._get_data_slice(
            [
                (axis.index_in_array, slice(None, axis.size)),
            ]
        )

        # Kinetic definitions
        ke = e0 * (1 + e0 / 2.0 / me) / (1 + e0 / me) ** 2
        tgt = e0 * (2 * me + e0) / (me + e0)
        rk0 = 2590 * (1 + e0 / me) * np.sqrt(2 * ke / me)

        for io in range(iterations):
            # Calculation of the ELF by normalization of the SSD
            # Norm(SSD) = Imag(-1/epsilon) (Energy Loss Function, ELF)

            # We start by the "angular corrections"
            Im = s.data / (np.log(1 + (beta * tgt / eaxis) ** 2)) / axis.scale
            if n is None and t is None:
                raise ValueError(
                    "The thickness and the refractive index are "
                    "not defined. Please provide one of them."
                )
            elif n is not None and t is not None:
                raise ValueError(
                    "Please provide the refractive index OR the "
                    "thickness information, not both"
                )
            elif n is not None:
                # normalize using the refractive index.
                K = (Im / eaxis).sum(
                    axis=axis.index_in_array, keepdims=True
                ) * axis.scale
                K = K / (np.pi / 2) / (1 - 1.0 / n**2)
                # K = (K / (np.pi / 2) / (1 - 1. / n ** 2)).reshape(
                #    np.insert(K.shape, axis.index_in_array, 1))
                # Calculate the thickness only if possible and required
                if zlp is not None and (full_output is True or iterations > 1):
                    te = 332.5 * K * ke / i0
                    if full_output is True:
                        output["thickness"] = te
            elif t is not None:
                if zlp is None:
                    raise ValueError(
                        "The ZLP must be provided when the  "
                        "thickness is used for normalization."
                    )
                # normalize using the thickness
                K = t * i0 / (332.5 * ke)
                te = t
            Im = Im / K

            # Kramers Kronig Transform:
            # We calculate KKT(Im(-1/epsilon))=1+Re(1/epsilon) with FFT
            # Follows: D W Johnson 1975 J. Phys. A: Math. Gen. 8 490
            # Use an optimal FFT size to speed up the calculation, and
            # make it double the closest upper value to workaround the
            # wrap-around problem.
            esize = optimal_fft_size(2 * axis.size)
            q = -2 * np.fft.fft(Im, esize, axis.index_in_array).imag / esize

            q[slicer] *= -1
            q = np.fft.fft(q, axis=axis.index_in_array)
            # Final touch, we have Re(1/eps)
            Re = q[slicer].real + 1

            # Egerton does this to correct the wrap-around problem, but in our
            # case this is not necessary because we compute the fft on an
            # extended and padded spectrum to avoid this problem.
            # Re=real(q)
            # Tail correction
            # vm=Re[axis.size-1]
            # Re[:(axis.size-1)]=Re[:(axis.size-1)]+1-(0.5*vm*((axis.size-1) /
            #  (axis.size*2-arange(0,axis.size-1)))**2)
            # Re[axis.size:]=1+(0.5*vm*((axis.size-1) /
            #  (axis.size+arange(0,axis.size)))**2)

            # Epsilon appears:
            #  We calculate the real and imaginary parts of the CDF
            e1 = Re / (Re**2 + Im**2)
            e2 = Im / (Re**2 + Im**2)

            if iterations > 1 and zlp is not None:
                # Surface losses correction:
                #  Calculates the surface ELF from a vacuum border effect
                #  A simulated surface plasmon is subtracted from the ELF
                Srfelf = 4 * e2 / ((e1 + 1) ** 2 + e2**2) - Im
                adep = tgt / (eaxis + delta) * np.arctan(
                    beta * tgt / axis.axis
                ) - beta / 1000.0 / (beta**2 + axis.axis**2.0 / tgt**2)
                Srfint = 2000 * K * adep * Srfelf / rk0 / te * axis.scale
                s.data = sorig.data - Srfint
                _logger.debug("Iteration number: %d / %d", io + 1, iterations)
                if iterations == io + 1 and full_output is True:
                    sp = sorig._deepcopy_with_new_data(Srfint)
                    sp.metadata.General.title += (
                        " estimated surface plasmon excitation."
                    )
                    output["surface plasmon estimation"] = sp
                    del sp
                del Srfint

        eps = s._deepcopy_with_new_data(e1 + e2 * 1j)
        del s
        eps.set_signal_type("DielectricFunction")
        eps.metadata.General.title = (
            self.metadata.General.title + "dielectric function "
            "(from Kramers-Kronig analysis)"
        )
        if eps.tmp_parameters.has_item("filename"):
            eps.tmp_parameters.filename = (
                self.tmp_parameters.filename + "_CDF_after_Kramers_Kronig_transform"
            )
        if "thickness" in output:
            # As above,prevent errors if the signal is a single spectrum
            if len(te) != 1:
                te = te[self.axes_manager._get_data_slice([(axis.index_in_array, 0)])]
            thickness = eps._get_navigation_signal(data=te)
            thickness.metadata.General.title = (
                self.metadata.General.title + " thickness "
                "(calculated using Kramers-Kronig analysis)"
            )
            output["thickness"] = thickness
        if full_output is False:
            return eps
        else:
            return eps, output

    def create_model(
        self,
        low_loss=None,
        auto_background=True,
        auto_add_edges=True,
        GOS="gosh",
        gos_file_path=None,
        dictionary=None,
    ):
        """Create a model for the current EELS data.

        Parameters
        ----------
        %s

        Returns
        -------
        model : :class:`~.models.eelsmodel.EELSModel` instance.

        Raises
        ------
        NotImplementedError
            If the signal axis is a non-uniform axis.
        """
        from exspy.models.eelsmodel import EELSModel

        if low_loss is not None and not self.axes_manager.signal_axes[0].is_uniform:
            raise NotImplementedError(
                "Multiple scattering is not implemented for spectra with a "
                "non-uniform energy axis. To create a model that does not "
                "account for multiple-scattering do not set the `ll` keyword."
            )
        model = EELSModel(
            self,
            low_loss=low_loss,
            auto_background=auto_background,
            auto_add_edges=auto_add_edges,
            GOS=GOS,
            dictionary=dictionary,
        )
        return model

    create_model.__doc__ %= EELSMODEL_PARAMETERS

    def plot(self, plot_edges=False, only_edges=("Major", "Minor"), **kwargs):
        """
        Plot the EELS spectrum. Markers indicating the position of the
        EELS edges can be added.

        Parameters
        ----------
        plot_edges : {False, True, list of string or string}
            If True, draws on s.metadata.Sample.elements for edges.
            Alternatively, provide a string of a single edge, or an iterable
            containing a list of valid elements, EELS families or edges. For
            example, an element should be 'Zr', an element edge family should
            be 'Zr_L' or an EELS edge 'Zr_L3'.
        only_edges : tuple of string
            Either 'Major' or 'Minor'. Defaults to both.
        kwargs
            The extra keyword arguments for plot()
        """

        super().plot(**kwargs)

        if plot_edges:
            # edges is a mapping {edge_name:edge_energy}
            edges = self._get_edges_to_plot(plot_edges, only_edges)
            self._plot_edge_labels(edges)

        self._plot.signal_plot.events.closed.connect(self._on_signal_plot_closing, [])

    def _on_signal_plot_closing(self):
        self._edge_markers = {"lines": None, "texts": None, "names": []}

    def _get_offsets_and_segments(self, edges):
        index = np.array([float(v) for v in edges.values()])  # dictionaries
        segments = np.empty((len(index), 2, 2))
        offsets = np.empty((len(index), 2))
        for i, ind in enumerate(index):
            segments[i] = [[ind, 1], [ind, 1.1]]
            offsets[i] = [ind, 1.1]
        return offsets, segments

    def _initialise_markers(self):
        self._edge_markers["lines"] = Lines(
            segments=np.empty((0, 2, 2)),
            transform="relative",
            color="black",
            shift=np.array([0.0, 0.19]),
        )
        self._edge_markers["texts"] = Texts(
            offsets=np.empty((0, 2)),
            texts=np.empty((0,)),
            offset_transform="relative",
            rotation=np.pi / 2,
            horizontalalignment="left",
            verticalalignment="bottom",
            facecolor="black",
            shift=0.2,
        )
        for key in ["lines", "texts"]:
            self.add_marker(self._edge_markers[key], render_figure=False)

    def _plot_edge_labels(self, edges):
        """
        Plot the EELS edge label (vertical line segment and text box) on
        the signal

        Parameters
        ----------
        edges : dictionary
            A dictionary with the labels as keys and their energies as values.
            For example, {'Fe_L2': 721.0, 'O_K': 532.0}

        """
        # the object is needed to connect replot method when axes_manager
        # indices changed
        _ = EdgesRange(self, interactive=False)
        self._add_edge_labels(edges)

    def _get_edges_to_plot(self, plot_edges, only_edges):
        # get the dictionary of the edge to be shown
        extra_element_edge_family = []
        if plot_edges is True:
            try:
                elements = self.metadata.Sample.elements
            except AttributeError:
                raise ValueError(
                    "No elements defined. Add them with "
                    "s.add_elements, or specify elements, edge "
                    "families or edges directly"
                )
        else:
            extra_element_edge_family.extend(np.atleast_1d(plot_edges))
            try:
                elements = self.metadata.Sample.elements
            except:
                elements = []

        element_edge_family = elements + extra_element_edge_family
        edges_dict = self._get_edges(element_edge_family, only_edges)

        return edges_dict

    def _get_edges(self, element_edge_family, only_edges):
        # get corresponding information depending on whether it is an element
        # a particular edge or a family of edge
        axis_min = self.axes_manager[-1].low_value
        axis_max = self.axes_manager[-1].high_value

        names_and_energies = {}
        shells = ["K", "L", "M", "N", "O"]

        errmsg = "Edge family '{}' is not supported. Supported edge family " "is {}."
        for member in element_edge_family:
            try:
                element, ss = member.split("_")

                if len(ss) == 1:
                    memtype = "family"
                    if ss not in shells:
                        raise AttributeError(errmsg.format(ss, shells))
                if len(ss) == 2:
                    memtype = "edge"
                    if ss[0] not in shells:
                        raise AttributeError(errmsg.format(ss[0], shells))
            except ValueError:
                element = member
                ss = ""
                memtype = "element"

            try:
                Binding_energies = elements_db[element]["Atomic_properties"][
                    "Binding_energies"
                ]
            except KeyError as err:
                raise ValueError("'{}' is not a valid element".format(element)) from err

            for edge in Binding_energies.keys():
                relevance = Binding_energies[edge]["relevance"]
                energy = Binding_energies[edge]["onset_energy (eV)"]

                isInRel = relevance in only_edges
                isInRng = axis_min < energy < axis_max
                isSameFamily = ss in edge

                if memtype == "element":
                    flag = isInRel & isInRng
                    edge_key = element + "_" + edge
                elif memtype == "edge":
                    flag = isInRng & (edge == ss)
                    edge_key = member
                elif memtype == "family":
                    flag = isInRel & isInRng & isSameFamily
                    edge_key = element + "_" + edge

                if flag:
                    names_and_energies[edge_key] = energy

        return names_and_energies

    def _remove_edge_labels(self, edge_names=None, render_figure=True):
        """
        Remove EELS edges markers to the signal

        Parameters
        ----------
        edge_names : str, list of str or None
            The string must be the name of edges, e. g. 'Fe_L2'.
            If ``None`` (default), remove all edges.
        render_figure : bool
            If True, render the figure after adding the markers
        """
        if edge_names is None:
            edge_names = self._edge_markers["names"]
        if isinstance(edge_names, set):
            # convert to list to find the index
            edge_names = list(edge_names)
        if not isinstance(edge_names, (list, tuple, np.ndarray)):
            edge_names = [edge_names]

        ind = np.where(np.isin(self._edge_markers["names"], edge_names))

        if self._edge_markers["lines"] is not None:
            self._edge_markers["lines"].remove_items(ind)
        if self._edge_markers["texts"] is not None:
            self._edge_markers["texts"].remove_items(ind)
        if self._edge_markers["names"] is not []:
            self._edge_markers["names"] = np.delete(self._edge_markers["names"], ind)

        if render_figure:
            self._render_figure(plot=["signal_plot"])

    def _add_edge_labels(self, edges, render_figure=True):
        """
        Add EELS edges markers to the signal

        Parameters
        ----------
        edge_name : dictionary or set
            If dictionary must be the name of edge as key and energy as values,
            e.g. {'Cr_L2': 584.0}. If list or set, must the name of the edge,
            e.g. set('Cr_L2', )
        render_figure : bool
            If True, render the figure after adding the markers
        """
        if isinstance(edges, set):
            edges_dict = {}
            for edge in edges:
                element, ss = edge.split("_")
                Binding_energies = elements_db[element]["Atomic_properties"][
                    "Binding_energies"
                ]
                edges_dict[edge] = Binding_energies[ss]["onset_energy (eV)"]
            edges = edges_dict

        offsets, segments = self._get_offsets_and_segments(edges)
        names = list(edges.keys())

        self._edge_markers["lines"].add_items(segments=segments)
        self._edge_markers["lines"].update()
        self._edge_markers["texts"].add_items(offsets=offsets, texts=names)
        self._edge_markers["lines"].update()
        self._edge_markers["names"] = np.append(self._edge_markers["names"], names)

        if render_figure:
            self._render_figure(plot=["signal_plot"])

    def _get_complementary_edges(self, edges, only_major=False):
        """
        Get other edges of the same element present within the energy
        range of the axis

        Parameters
        ----------
        edges : iterable
            A sequence of strings contains edges in the format of
            element_subshell for EELS. For example, ['Fe_L2', 'O_K']
        only_major : bool
            Whether to show only the major edges. The default is False.

        Returns
        -------
        complmt_edges : list
            A list containing all the complementary edges of the same element
            present within the energy range of the axis
        """

        emin = self.axes_manager[-1].low_value
        emax = self.axes_manager[-1].high_value
        complmt_edges = []

        elements = set()
        for edge in edges:
            element, _ = edge.split("_")
            elements.update([element])

        for element in elements:
            ss_info = elements_db[element]["Atomic_properties"]["Binding_energies"]

            for subshell in ss_info:
                sse = ss_info[subshell]["onset_energy (eV)"]
                ssr = ss_info[subshell]["relevance"]

                if only_major:
                    if ssr != "Major":
                        continue

                edge = element + "_" + subshell
                if (
                    (emin <= sse <= emax)
                    and (subshell[-1] != "a")
                    and (edge not in edges)
                ):
                    complmt_edges.append(edge)

        return complmt_edges

    def rebin(self, new_shape=None, scale=None, crop=True, dtype=None, out=None):
        factors = self._validate_rebin_args_and_get_factors(
            new_shape=new_shape, scale=scale
        )
        m = super().rebin(
            new_shape=new_shape, scale=scale, crop=crop, dtype=dtype, out=out
        )
        m = out or m
        time_factor = np.prod(
            [factors[axis.index_in_array] for axis in m.axes_manager.navigation_axes]
        )
        mdeels = m.metadata
        m.get_dimensions_from_data()
        if m.metadata.get_item("Acquisition_instrument.TEM.Detector.EELS"):
            mdeels = m.metadata.Acquisition_instrument.TEM.Detector.EELS
            if "dwell_time" in mdeels:
                mdeels.dwell_time *= time_factor
            if "exposure" in mdeels:
                mdeels.exposure *= time_factor
        else:
            _logger.info(
                "No dwell_time could be found in the metadata so "
                "this has not been updated."
            )
        if out is None:
            return m
        else:
            out.events.data_changed.trigger(obj=out)
        return m

    rebin.__doc__ = hyperspy.signal.BaseSignal.rebin.__doc__

    def vacuum_mask(
        self, threshold=10.0, start_energy=None, closing=True, opening=False
    ):
        """
        Generate mask of the vacuum region

        Parameters
        ----------
        threshold: float
            For a given navigation coordinate, mean value in the energy axis
            below which the pixel is considered as vacuum.
        start_energy: float, None
            Minimum energy included in the calculation of the mean intensity.
            If None, consider only the last quarter of the spectrum to
            calculate the mask.
        closing: bool
            If True, a morphological closing is applied to the mask.
        opening: bool
            If True, a morphological opening is applied to the mask.

        Returns
        -------
        mask: signal
            The mask of the region.
        """
        if self.axes_manager.navigation_dimension == 0:
            raise RuntimeError(
                "Navigation dimenstion must be higher than 0 "
                "to estimate a vacuum mask."
            )
        signal_axis = self.axes_manager.signal_axes[0]
        if start_energy is None:
            start_energy = 0.75 * signal_axis.high_value

        mask = self.isig[start_energy:].mean(-1) <= threshold

        from scipy.ndimage import binary_dilation, binary_erosion

        if closing:
            mask.data = binary_dilation(mask.data, border_value=0)
            mask.data = binary_erosion(mask.data, border_value=1)
        if opening:
            mask.data = binary_erosion(mask.data, border_value=1)
            mask.data = binary_dilation(mask.data, border_value=0)
        return mask


class LazyEELSSpectrum(EELSSpectrum, LazySignal1D):

    """Lazy signal class for EELS spectra."""

    __doc__ += LAZYSIGNAL_DOC.replace("__BASECLASS__", "EELSSpectrum")



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The exSpy developers
#
# This file is part of exSpy.
#
# exSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

import configparser
import logging
import os
from pathlib import Path

import traits.api as t
from hyperspy.ui_registry import add_gui_method


config_path = Path("~/.exspy").expanduser()
config_path.mkdir(parents=True, exist_ok=True)
defaults_file = Path(config_path, "exspyrc")

_logger = logging.getLogger(__name__)


def guess_gos_path():
    if os.name in ["nt", "dos"]:
        # If DM is installed, use the GOS tables from the default
        # installation
        # location in windows
        program_files = os.environ["PROGRAMFILES"]
        gos = "Gatan\\DigitalMicrograph\\EELS Reference Data\\H-S GOS Tables"
        gos_path = Path(program_files, gos)

        # Else, use the default location in the .hyperspy forlder
        if not gos_path.is_dir() and "PROGRAMFILES(X86)" in os.environ:
            program_files = os.environ["PROGRAMFILES(X86)"]
            gos_path = Path(program_files, gos)
            if not gos_path.is_dir():
                gos_path = Path(config_path, "EELS_GOS")
    else:
        gos_path = Path(config_path, "EELS_GOS")
    return gos_path


class EELSConfig(t.HasTraits):
    eels_gos_files_path = t.Directory(
        guess_gos_path(),
        label="Hartree-Slater GOS directory",
        desc="The GOS files are used to create the EELS edge components",
    )


class EDSConfig(t.HasTraits):
    eds_mn_ka = t.CFloat(
        130.0,
        label="Energy resolution at Mn Ka (eV)",
        desc="default value for FWHM of the Mn Ka peak in eV,"
        "This value is used as a first approximation"
        "of the energy resolution of the detector.",
    )
    eds_tilt_stage = t.CFloat(
        0.0, label="Stage tilt", desc="default value for the stage tilt in degree."
    )
    eds_detector_azimuth = t.CFloat(
        0.0,
        label="Azimuth angle",
        desc="default value for the azimuth angle in degree. If the azimuth"
        " is zero, the detector is perpendicular to the tilt axis.",
    )
    eds_detector_elevation = t.CFloat(
        35.0,
        label="Elevation angle",
        desc="default value for the elevation angle in degree.",
    )


template = {
    "EELS": EELSConfig(),
    "EDS": EDSConfig(),
}


def template2config(template, config):
    for section, traited_class in template.items():
        config.add_section(section)
        for key, item in traited_class.trait_get().items():
            config.set(section, key, str(item))


def config2template(template, config):
    for section, traited_class in template.items():
        config_dict = {}
        for name, value in config.items(section):
            if value == "True":
                value = True
            elif value == "False":
                value = False
            config_dict[name] = value
        traited_class.trait_set(True, **config_dict)


defaults_file_exists = defaults_file.is_file()

config = configparser.ConfigParser(allow_no_value=True)
template2config(template, config)
rewrite = False
if defaults_file_exists:
    # Parse the config file. It only copy to config the options that are
    # already defined. If the file contains any option that was not already
    # define the config file is rewritten because it is obsolate

    config2 = configparser.ConfigParser(allow_no_value=True)
    config2.read(defaults_file)
    for section in config2.sections():
        if config.has_section(section):
            for option in config2.options(section):
                if config.has_option(section, option):
                    config.set(section, option, config2.get(section, option))
                else:
                    rewrite = True
        else:
            rewrite = True

if not defaults_file_exists or rewrite is True:
    _logger.info("Writing the config file")
    with open(defaults_file, "w") as df:
        config.write(df)

# Use the traited classes to cast the content of the ConfigParser
config2template(template, config)


@add_gui_method(toolkey="exspy.Preferences")
class Preferences(t.HasTraits):
    EELS = t.Instance(EELSConfig)
    EDS = t.Instance(EDSConfig)

    def save(self):
        config = configparser.ConfigParser(allow_no_value=True)
        template2config(template, config)
        config.write(open(defaults_file, "w"))


preferences = Preferences(
    EELS=template["EELS"],
    EDS=template["EDS"],
)



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The exSpy developers
#
# This file is part of exSpy.
#
# exSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

import pytest

from exspy.misc.eels.tools import get_edges_near_energy, get_info_from_edges


def test_single_edge():
    edges = get_edges_near_energy(532, width=0)
    assert len(edges) == 1
    assert edges == ["O_K"]


def test_multiple_edges():
    edges = get_edges_near_energy(640, width=100)
    assert len(edges) == 16
    assert edges == [
        "Mn_L3",
        "Ra_N4",
        "I_M4",
        "Cd_M2",
        "Mn_L2",
        "V_L1",
        "I_M5",
        "Cd_M3",
        "In_M3",
        "Xe_M5",
        "Ac_N4",
        "Ra_N5",
        "Fr_N4",
        "Ag_M2",
        "F_K",
        "Xe_M4",
    ]


def test_multiple_edges_ascending():
    edges = get_edges_near_energy(640, width=100, order="ascending")
    assert len(edges) == 16
    assert edges == [
        "Ag_M2",
        "Ra_N5",
        "Fr_N4",
        "Cd_M3",
        "I_M5",
        "V_L1",
        "I_M4",
        "Ra_N4",
        "Mn_L3",
        "Cd_M2",
        "Mn_L2",
        "In_M3",
        "Xe_M5",
        "Ac_N4",
        "F_K",
        "Xe_M4",
    ]


def test_multiple_edges_descending():
    edges = get_edges_near_energy(640, width=100, order="descending")
    assert len(edges) == 16
    assert edges == [
        "F_K",
        "Xe_M4",
        "Ac_N4",
        "Xe_M5",
        "In_M3",
        "Cd_M2",
        "Mn_L2",
        "Mn_L3",
        "Ra_N4",
        "I_M4",
        "V_L1",
        "I_M5",
        "Cd_M3",
        "Ra_N5",
        "Fr_N4",
        "Ag_M2",
    ]


def test_negative_energy_width():
    with pytest.raises(Exception):
        get_edges_near_energy(849, width=-5)


def test_wrong_ordering():
    with pytest.raises(ValueError):
        get_edges_near_energy(532, order="random")


def test_info_one_edge():
    info = get_info_from_edges("O_K")
    assert len(info) == 1


def test_info_multiple_edges():
    info = get_info_from_edges(["O_K", "N_K", "Cr_L3"])
    assert len(info) == 3


def test_info_wrong_edge_format():
    with pytest.raises(ValueError):
        get_info_from_edges(["O_K", "NK"])



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The exSpy developers
#
# This file is part of exSpy.
#
# exSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exSpy. If not, see <https://www.gnu.org/licenses/#GPL>.


import functools
import logging
import math

import numpy as np
from scipy.interpolate import splev

from hyperspy.component import Component
from exspy.misc.eels.gosh_gos import GoshGOS, _GOSH_DOI
from exspy.misc.eels.hartree_slater_gos import HartreeSlaterGOS
from exspy.misc.eels.hydrogenic_gos import HydrogenicGOS
from exspy.misc.eels.effective_angle import effective_angle
from hyperspy.ui_registry import add_gui_method


_logger = logging.getLogger(__name__)


class FSet(set):
    def __init__(self, component, *args, **kwargs):
        """Creates a set that knows about the Component

        Parameters:
        -----------
        component : Component
            The component to which the FSet belongs.
        """
        self.component = component
        super().__init__(*args, **kwargs)

    @functools.wraps(set.add)
    def add(self, item):
        item.active = self.component.fine_structure_active
        if self.component.model and item not in self.component.model:
            self.component.model.append(item)
        super().add(item)

    @functools.wraps(set.update)
    def update(self, iterable):
        for item in iterable:
            item.active = self.component.fine_structure_active
            if self.component.model and item not in self.component.model:
                self.component.model.append(item)
        super().update(iterable)


@add_gui_method(toolkey="exspy.EELSCLEdge_Component")
class EELSCLEdge(Component):
    """
    EELS core loss ionisation edge from hydrogenic or tabulated
    GOS with splines for fine structure fitting.

    Hydrogenic GOS are limited to K and L shells.

    Several possibilities are available for tabulated GOS.

    The preferred option is to use a database of cross sections in GOSH
    format. One such database can be freely downloaded from Zenodo at:
    https://zenodo.org/record/6599071 while information on the GOSH format
    are available at: https://gitlab.com/gguzzina/gosh .

    exSpy also supports Peter Rez's Hartree Slater cross sections
    parametrised as distributed by Gatan in their Digital Micrograph (DM)
    software. If Digital Micrograph is installed in the system exSpy in the
    standard location exSpy should find the path to the HS GOS folder.
    Otherwise, the location of the folder can be defined in exSpy
    preferences, which can be done through ~:func:`~.api.preferences.gui` or
    the :attr:`~api.preferences.EELS.eels_gos_files_path` variable.

    Parameters
    ----------
    element_subshell : str or dict
        Usually a string, for example, ``'Ti_L3'`` for the GOS of the titanium L3
        subshell. If a dictionary is passed, it is assumed that Hartree Slater
        GOS was exported using `GOS.as_dictionary`, and will be reconstructed.
    GOS : ``'gosh'``, ``'hydrogenic'``, ``'Hartree-Slater'`` or str
        The GOS to use. Default is ``'gosh'``. If str, it must the path to gosh
        GOS file.
    gos_file_path : str, None
        Only with GOS='gosh'. Specify the file path of the gosh file
        to use. If None, use the file from doi:{}

    Attributes
    ----------
    onset_energy : Parameter
        The edge onset position
    intensity : Parameter
        The factor by which the cross section is multiplied, what in
        favourable cases is proportional to the number of atoms of
        the element. It is a component.Parameter instance.
        It is fixed by default.
    effective_angle : Parameter
        The effective collection semi-angle. It is automatically
        calculated by ``set_microscope_parameters``. It is a
        component.Parameter instance. It is fixed by default.
    fine_structure_active : bool, default False
        Activates/deactivates the fine structure features. When active,
        the fine structure region
        is not defined by the simulated
        EELS core-loss edge, but by a spline (if ``fine_structure_spline_active``
        is ``True``) and/or any component in ``fine_structure_components``.
    fine_structure_spline_active : bool, default True
        If True and ``fine_structure_active`` is True, the region from
        ``fine_structure_spline_onset`` until ``fine_structure_width``
        are modelled with a spline.
    fine_structure_coeff : Parameter
        The coefficients of the spline that fits the fine structure.
        Fix this parameter to fix the fine structure. It is a
        ``component.Parameter`` instance.
    fine_structure_smoothing : float between 0 and 1, default 0.3
        Controls the level of smoothing of the fine structure model.
        Decreasing the value increases the level of smoothing.
    fine_structure_spline_onset : float, default 0.
        The position, from the ionization edge onset, at which the region
        modelled by the spline function starts.
    fine_structure_width : float, default 30.
        The width of the energy region, from the ionization edge onset, where
        the model is a spline function and/or any component
        in ``fine_structure_components`` instead of the EELS ionization
        edges simulation.
    fine_structure_components : set, default ``set()``
        A set containing components to model the fine structure region
        of the EELS ionization edge.
    """.format(
        _GOSH_DOI
    )

    _fine_structure_smoothing = 0.3
    _fine_structure_coeff_free = True
    _fine_structure_spline_active = True

    def __init__(self, element_subshell, GOS="gosh", gos_file_path=None):
        # Declare the parameters
        self.fine_structure_components = FSet(component=self)
        Component.__init__(
            self,
            ["intensity", "fine_structure_coeff", "effective_angle", "onset_energy"],
            linear_parameter_list=["intensity"],
        )
        if isinstance(element_subshell, dict):
            self.element = element_subshell["element"]
            self.subshell = element_subshell["subshell"]
        else:
            self.element, self.subshell = element_subshell.split("_")
        self.name = "_".join([self.element, self.subshell])
        self.energy_scale = None
        self.effective_angle.free = False
        self.fine_structure_active = False
        self.fine_structure_width = 30.0
        self.fine_structure_coeff.ext_force_positive = False
        self.GOS = None

        if GOS == "gosh":
            self.GOS = GoshGOS(element_subshell, gos_file_path=gos_file_path)
        elif GOS == "Hartree-Slater":  # pragma: no cover
            self.GOS = HartreeSlaterGOS(element_subshell)
        elif GOS == "hydrogenic":
            self.GOS = HydrogenicGOS(element_subshell)
        else:
            raise ValueError(
                "GOS must be one of 'gosh', 'hydrogenic' or 'Hartree-Slater'."
            )
        self.onset_energy.value = self.GOS.onset_energy
        self.onset_energy.free = False
        self._position = self.onset_energy
        self.free_onset_energy = False
        self.intensity.grad = self.grad_intensity
        self.intensity.value = 1
        self.intensity.bmin = 0.0
        self.intensity.bmax = None

        self._whitelist["GOS"] = ("init", GOS)
        if GOS == "gosh":
            self._whitelist["element_subshell"] = ("init", self.GOS.as_dictionary(True))
        elif GOS == "Hartree-Slater":  # pragma: no cover
            self._whitelist["element_subshell"] = ("init", self.GOS.as_dictionary(True))
        elif GOS == "hydrogenic":
            self._whitelist["element_subshell"] = ("init", element_subshell)
        self._whitelist["fine_structure_active"] = None
        self._whitelist["fine_structure_width"] = None
        self._whitelist["fine_structure_smoothing"] = None
        self._whitelist["fine_structure_spline_onset"] = None
        self._whitelist["fine_structure_spline_active"] = None
        self._whitelist["_fine_structure_coeff_free"] = None
        self.effective_angle.events.value_changed.connect(self._integrate_GOS, [])
        self.onset_energy.events.value_changed.connect(self._integrate_GOS, [])
        self.onset_energy.events.value_changed.connect(self._calculate_knots, [])
        self._fine_structure_spline_onset = 0
        self.events.active_changed.connect(self._set_active_fine_structure_components)

    # Automatically fix the fine structure when the fine structure is
    # disable.
    # In this way we avoid a common source of problems when fitting
    # However the fine structure must be *manually* freed when we
    # reactivate the fine structure.
    @property
    def fine_structure_active(self):
        return self.__fine_structure_active

    @fine_structure_active.setter
    def fine_structure_active(self, arg):
        if self.fine_structure_spline_active:
            if arg:
                self.fine_structure_coeff.free = self._fine_structure_coeff_free
            else:
                self._fine_structure_coeff_free = self.fine_structure_coeff.free
                self.fine_structure_coeff.free = False
        for comp in self.fine_structure_components:
            if isinstance(comp, str):
                # Loading from a dictionary and the external fine structure components are still strings
                break
            comp.active = arg
        self.__fine_structure_active = arg
        if self.fine_structure_spline_active and self.model:
            self.model.update_plot()

    @property
    def fine_structure_width(self):
        return self.__fine_structure_width

    @fine_structure_width.setter
    def fine_structure_width(self, arg):
        self.__fine_structure_width = arg
        self._set_fine_structure_coeff()
        if self.fine_structure_active and self.model:
            self.model.update_plot()

    # E0
    @property
    def E0(self):
        return self.__E0

    @E0.setter
    def E0(self, arg):
        self.__E0 = arg
        self._calculate_effective_angle()

    @property
    def collection_angle(self):
        return self.__collection_angle

    @collection_angle.setter
    def collection_angle(self, arg):
        self.__collection_angle = arg
        self._calculate_effective_angle()

    @property
    def convergence_angle(self):
        return self.__convergence_angle

    @convergence_angle.setter
    def convergence_angle(self, arg):
        self.__convergence_angle = arg
        self._calculate_effective_angle()

    def _calculate_effective_angle(self):
        try:
            self.effective_angle.value = effective_angle(
                self.E0,
                self.GOS.onset_energy,
                self.convergence_angle,
                self.collection_angle,
            )
        except BaseException:
            # All the parameters may not be defined yet...
            pass

    def _set_active_fine_structure_components(self, active, **kwargs):
        if not self.fine_structure_active:
            return
        for comp in self.fine_structure_components:
            comp.active = active

    @property
    def fine_structure_smoothing(self):
        """Controls the level of the smoothing of the fine structure.

        It must a real number between 0 and 1. The higher close to 0
        the higher the smoothing.

        """
        return self._fine_structure_smoothing

    @fine_structure_smoothing.setter
    def fine_structure_smoothing(self, value):
        if 0 <= value <= 1:
            self._fine_structure_smoothing = value
            self._set_fine_structure_coeff()
            if self.fine_structure_active and self.model:
                self.model.update_plot()
        else:
            raise ValueError("The value must be a number between 0 and 1")

    # It is needed because the property cannot be used to sort the edges
    def _onset_energy(self):
        return self.onset_energy.value

    @property
    def fine_structure_spline_onset(self):
        return self._fine_structure_spline_onset

    @fine_structure_spline_onset.setter
    def fine_structure_spline_onset(self, value):
        if not np.allclose(value, self._fine_structure_spline_onset):
            self._fine_structure_spline_onset = value
            self._set_fine_structure_coeff()
            if self.fine_structure_active and self.model:
                self.model.update_plot()

    @property
    def fine_structure_spline_active(self):
        return self._fine_structure_spline_active

    @fine_structure_spline_active.setter
    def fine_structure_spline_active(self, value):
        if value:
            if self.fine_structure_active:
                self.fine_structure_coeff.free = self._fine_structure_coeff_free
            self._fine_structure_spline_active = value
        else:
            self._fine_structure_coeff_free = self.fine_structure_coeff.free
            self.fine_structure_coeff.free = False
            self._fine_structure_spline_active = value
        if self.fine_structure_active and self.model:
            self.model.update_plot()

    def _set_fine_structure_coeff(self):
        if self.energy_scale is None:
            return
        self.fine_structure_coeff._number_of_elements = (
            int(
                round(
                    self.fine_structure_smoothing
                    * (self.fine_structure_width - self.fine_structure_spline_onset)
                    / self.energy_scale
                )
            )
            + 4
        )
        self.fine_structure_coeff.bmin = None
        self.fine_structure_coeff.bmax = None
        self._calculate_knots()
        if self.fine_structure_coeff.map is not None:
            self.fine_structure_coeff._create_array()

    def fix_fine_structure(self):
        """Fixes the fine structure spline and the parameters of the fine
        structure components, if any.

        See Also
        --------
        free_fine_structure

        """
        self.fine_structure_coeff.free = False
        for component in self.fine_structure_components:
            component._auto_free_parameters = []
            for parameter in component.free_parameters:
                parameter.free = False
                component._auto_free_parameters.append(parameter)

    def free_fine_structure(self):
        """Frees the parameters of the fine structure

        If there are fine structure components, only the
        parameters that have been previously fixed with
        ``fix_fine_structure`` will be set free.

        The spline parameters set free only if
        ``fine_structure_spline_active`` is ``True``.

        See Also
        --------
        free_fine_structure

        """
        if self.fine_structure_spline_active:
            self.fine_structure_coeff.free = True
        for component in self.fine_structure_components:
            if hasattr(component, "_auto_free_parameters"):
                for parameter in component._auto_free_parameters:
                    parameter.free = True
                del component._auto_free_parameters

    def set_microscope_parameters(self, E0, alpha, beta, energy_scale):
        """
        Set the microscope parameters.

        Parameters
        ----------
        E0 : float
            Electron beam energy in keV.
        alpha: float
            Convergence semi-angle in mrad.
        beta: float
            Collection semi-angle in mrad.
        energy_scale : float
            The energy step in eV.
        """
        # Relativistic correction factors
        old = self.effective_angle.value
        with self.effective_angle.events.value_changed.suppress_callback(
            self._integrate_GOS
        ):
            self.convergence_angle = alpha
            self.collection_angle = beta
            self.energy_scale = energy_scale
            self.E0 = E0
        if self.effective_angle.value != old:
            self._integrate_GOS()

    def _integrate_GOS(self):
        # Integration over q using splines
        angle = self.effective_angle.value * 1e-3  # in rad
        self.tab_xsection = self.GOS.integrateq(self.onset_energy.value, angle, self.E0)
        # Calculate extrapolation powerlaw extrapolation parameters
        E1 = self.GOS.energy_axis[-2] + self.GOS.energy_shift
        E2 = self.GOS.energy_axis[-1] + self.GOS.energy_shift
        y1 = self.GOS.qint[-2]  # in m**2/bin */
        y2 = self.GOS.qint[-1]  # in m**2/bin */
        self._power_law_r = math.log(y2 / y1) / math.log(E1 / E2)
        self._power_law_A = y1 / E1**-self._power_law_r

    def _calculate_knots(self):
        start = self.onset_energy.value
        stop = start + self.fine_structure_width
        self.__knots = np.r_[
            [start] * 4,
            np.linspace(start, stop, self.fine_structure_coeff._number_of_elements)[
                2:-2
            ],
            [stop] * 4,
        ]

    def function(self, E):
        """Returns the number of counts in barns"""
        shift = self.onset_energy.value - self.GOS.onset_energy
        if shift != self.GOS.energy_shift:
            # Because hspy Events are not executed in any given order,
            # an external function could be in the same event execution list
            # as _integrate_GOS and be executed first. That can potentially
            # cause an error that enforcing _integrate_GOS here prevents. Note
            # that this is suboptimal because _integrate_GOS is computed twice
            # unnecessarily.
            self._integrate_GOS()
        Emax = self.GOS.energy_axis[-1] + self.GOS.energy_shift
        cts = np.zeros_like(E, dtype="float")
        if self.fine_structure_active:
            ifsx1 = self.onset_energy.value + self.fine_structure_spline_onset
            ifsx2 = self.onset_energy.value + self.fine_structure_width
            if self.fine_structure_spline_active:
                bifs = (E >= ifsx1) & (E < ifsx2)
                # Only set the spline values if the spline is in the energy region
                if np.any(bifs):
                    cts[bifs] = splev(
                        E[bifs],
                        (self.__knots, self.fine_structure_coeff.value + (0,) * 4, 3),
                    )
            # The cross-section is set to 0 in the fine structure region
            itab = (E < Emax) & (E >= ifsx2)
        else:
            itab = (E < Emax) & (E >= self.onset_energy.value)
        if itab.any():
            cts[itab] = self.tab_xsection(E[itab])
        bext = E >= Emax
        if bext.any():
            cts[bext] = self._power_law_A * E[bext] ** -self._power_law_r
        return cts * self.intensity.value

    def grad_intensity(self, E):
        return self.function(E) / self.intensity.value

    def fine_structure_coeff_to_txt(self, filename):
        np.savetxt(filename + ".dat", self.fine_structure_coeff.value, fmt="%12.6G")

    def txt_to_fine_structure_coeff(self, filename):
        fs = np.loadtxt(filename)
        self._calculate_knots()
        if len(fs) == len(self.__knots):
            self.fine_structure_coeff.value = fs
        else:
            raise ValueError(
                "The provided fine structure file "
                "doesn't match the size of the current fine structure"
            )

    def get_fine_structure_as_signal1D(self):
        """
        Returns a spectrum containing the fine structure.

        Notes
        -----
        The fine structure is corrected from multiple scattering if
        the model was convolved with a low-loss spectrum

        """
        from exspy.signals.eels import EELSSpectrum

        channels = int(np.floor(self.fine_structure_width / self.energy_scale))
        data = np.zeros(self.fine_structure_coeff.map.shape + (channels,))
        s = EELSSpectrum(data, axes=self.intensity._axes_manager._get_axes_dicts())
        s.get_dimensions_from_data()
        s.axes_manager.signal_axes[0].offset = self.onset_energy.value
        # Backup the axes_manager
        original_axes_manager = self._axes_manager
        self._axes_manager = s.axes_manager
        for spectrum in s:
            self.fetch_stored_values()
            spectrum.data[:] = self.function(s.axes_manager.signal_axes[0].axis)
        # Restore the axes_manager and the values
        self._axes_manager = original_axes_manager
        self.fetch_stored_values()

        s.metadata.General.title = self.name.replace("_", " ") + " fine structure"

        return s

    def as_dictionary(self, fullcopy=True):
        dic = super().as_dictionary(fullcopy=fullcopy)
        dic["fine_structure_components"] = [
            t.name for t in self.fine_structure_components
        ]
        dic["_whitelist"]["fine_structure_components"] = ""
        return dic



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The exSpy developers
#
# This file is part of exSpy.
#
# exSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

from importlib.metadata import version
from pathlib import Path

from . import components
from . import data
from . import models
from . import signals
from .misc import material
from ._defaults_parser import preferences


__version__ = version("exspy")

# For development version, `setuptools_scm` will be used at build time
# to get the dev version, in case of missing vcs information (git archive,
# shallow repository), the fallback version defined in pyproject.toml will
# be used

# If we have an editable installed from a git repository try to use
# `setuptools_scm` to find a more accurate version:
# `importlib.metadata` will provide the version at installation
# time and for editable version this may be different

# we only do that if we have enough git history, e.g. not shallow checkout
_root = Path(__file__).resolve().parents[1]
if (_root / ".git").exists() and not (_root / ".git/shallow").exists():
    try:
        # setuptools_scm may not be installed
        from setuptools_scm import get_version

        __version__ = get_version(_root)
    except ImportError:  # pragma: no cover
        # setuptools_scm not installed, we keep the existing __version__
        pass


__all__ = [
    "__version__",
    "components",
    "data",
    "preferences",
    "material",
    "models",
    "signals",
]


def __dir__():
    return sorted(__all__)



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The exSpy developers
#
# This file is part of exSpy.
#
# exSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

import itertools
import logging

import numpy as np
import warnings
from collections.abc import Iterable
from matplotlib import pyplot as plt

from hyperspy import utils
from hyperspy.signal import BaseSignal
from hyperspy._signals.signal1d import Signal1D, LazySignal1D
from exspy.misc.elements import elements as elements_db
from exspy.misc.eds import utils as utils_eds
from hyperspy.misc.utils import isiterable
from hyperspy.utils.markers import Texts, VerticalLines, Lines
from hyperspy.docstrings.plot import BASE_PLOT_DOCSTRING_PARAMETERS, PLOT1D_DOCSTRING
from hyperspy.docstrings.signal import LAZYSIGNAL_DOC

_logger = logging.getLogger(__name__)


class EDSSpectrum(Signal1D):
    """General signal class for EDS spectra."""

    _signal_type = "EDS"

    def __init__(self, *args, **kwards):
        super().__init__(*args, **kwards)
        if self.metadata.Signal.signal_type == "EDS":
            warnings.warn(
                "The microscope type is not set. Use "
                "set_signal_type('EDS_TEM')  "
                "or set_signal_type('EDS_SEM')"
            )
        self.axes_manager.signal_axes[0].is_binned = True
        self._xray_markers = {}

    def _get_line_energy(self, Xray_line, FWHM_MnKa=None):
        """
        Get the line energy and the energy resolution of a Xray line.

        The return values are in the same units than the signal axis

        Parameters
        ----------
        Xray_line : strings
            Valid element X-ray lines e.g. Fe_Kb
        FWHM_MnKa: {None, float, 'auto'}
            The energy resolution of the detector in eV
            if 'auto', used the one in
            'self.metadata.Acquisition_instrument.SEM.Detector.EDS.energy_resolution_MnKa'

        Returns
        -------
        float: the line energy, if FWHM_MnKa is None
        (float,float): the line energy and the energy resolution, if FWHM_MnKa
        is not None
        """

        units_name = self.axes_manager.signal_axes[0].units

        if FWHM_MnKa == "auto":
            if self.metadata.Signal.signal_type == "EDS_SEM":
                FWHM_MnKa = (
                    self.metadata.Acquisition_instrument.SEM.Detector.EDS.energy_resolution_MnKa
                )
            elif self.metadata.Signal.signal_type == "EDS_TEM":
                FWHM_MnKa = (
                    self.metadata.Acquisition_instrument.TEM.Detector.EDS.energy_resolution_MnKa
                )
            else:
                raise NotImplementedError(
                    "This method only works for EDS_TEM or EDS_SEM signals. "
                    "You can use `set_signal_type('EDS_TEM')` or"
                    "`set_signal_type('EDS_SEM')` to convert to one of these"
                    "signal types."
                )
        line_energy = utils_eds._get_energy_xray_line(Xray_line)
        if units_name == "eV":
            line_energy *= 1000
            if FWHM_MnKa is not None:
                line_FWHM = (
                    utils_eds.get_FWHM_at_Energy(FWHM_MnKa, line_energy / 1000) * 1000
                )
        elif units_name == "keV":
            if FWHM_MnKa is not None:
                line_FWHM = utils_eds.get_FWHM_at_Energy(FWHM_MnKa, line_energy)
        else:
            raise ValueError(
                f"{units_name} is not a valid units for the energy axis. "
                "Only `eV` and `keV` are supported. "
                "If `s` is the variable containing this EDS spectrum:\n "
                ">>> s.axes_manager.signal_axes[0].units = 'keV' \n"
            )
        if FWHM_MnKa is None:
            return line_energy
        else:
            return line_energy, line_FWHM

    def _get_beam_energy(self):
        """
        Get the beam energy.

        The return value is in the same units than the signal axis
        """

        if "Acquisition_instrument.SEM.beam_energy" in self.metadata:
            beam_energy = self.metadata.Acquisition_instrument.SEM.beam_energy
        elif "Acquisition_instrument.TEM.beam_energy" in self.metadata:
            beam_energy = self.metadata.Acquisition_instrument.TEM.beam_energy
        else:
            raise AttributeError(
                "The beam energy is not defined in `metadata`. "
                "Use `set_microscope_parameters` to set it."
            )

        units_name = self.axes_manager.signal_axes[0].units

        if units_name == "eV":
            beam_energy *= 1000
        return beam_energy

    def _get_xray_lines_in_spectral_range(self, xray_lines):
        """
        Return the lines in the energy range

        Parameters
        ----------
        xray_lines: List of string
            The xray_lines

        Return
        ------
        The list of xray_lines in the energy range
        """
        ax = self.axes_manager.signal_axes[0]
        low_value = ax.low_value
        high_value = ax.high_value
        try:
            if self._get_beam_energy() < high_value:
                high_value = self._get_beam_energy()
        except AttributeError:
            # in case the beam energy is not defined in the metadata
            pass
        xray_lines_in_range = []
        xray_lines_not_in_range = []
        for xray_line in xray_lines:
            line_energy = self._get_line_energy(xray_line)
            if low_value < line_energy < high_value:
                xray_lines_in_range.append(xray_line)
            else:
                xray_lines_not_in_range.append(xray_line)
        return xray_lines_in_range, xray_lines_not_in_range

    def sum(self, axis=None, out=None, rechunk=False):
        if axis is None:
            axis = self.axes_manager.navigation_axes
        s = super().sum(axis=axis, out=out, rechunk=rechunk)
        s = out or s

        # Update live time by the change in navigation axes dimensions
        time_factor = np.prod(
            [ax.size for ax in self.axes_manager.navigation_axes]
        ) / np.prod([ax.size for ax in s.axes_manager.navigation_axes])
        aimd = s.metadata.get_item("Acquisition_instrument", None)
        if aimd is not None:
            aimd = s.metadata.Acquisition_instrument
            if "SEM.Detector.EDS.live_time" in aimd:
                aimd.SEM.Detector.EDS.live_time *= time_factor
            elif "TEM.Detector.EDS.live_time" in aimd:
                aimd.TEM.Detector.EDS.live_time *= time_factor
            else:
                _logger.info(
                    "Live_time could not be found in the metadata and "
                    "has not been updated."
                )

        if out is None:
            return s

    sum.__doc__ = Signal1D.sum.__doc__

    def rebin(self, new_shape=None, scale=None, crop=True, dtype=None, out=None):
        factors = self._validate_rebin_args_and_get_factors(
            new_shape=new_shape,
            scale=scale,
        )
        m = super().rebin(
            new_shape=new_shape, scale=scale, crop=crop, dtype=dtype, out=out
        )
        m = out or m
        time_factor = np.prod(
            [factors[axis.index_in_array] for axis in m.axes_manager.navigation_axes]
        )
        aimd = m.metadata.Acquisition_instrument
        if "Acquisition_instrument.SEM.Detector.EDS.real_time" in m.metadata:
            aimd.SEM.Detector.EDS.real_time *= time_factor
        elif "Acquisition_instrument.TEM.Detector.EDS.real_time" in m.metadata:
            aimd.TEM.Detector.EDS.real_time *= time_factor
        else:
            _logger.info(
                "real_time could not be found in the metadata and has not been updated."
            )
        if "Acquisition_instrument.SEM.Detector.EDS.live_time" in m.metadata:
            aimd.SEM.Detector.EDS.live_time *= time_factor
        elif "Acquisition_instrument.TEM.Detector.EDS.live_time" in m.metadata:
            aimd.TEM.Detector.EDS.live_time *= time_factor
        else:
            _logger.info(
                "Live_time could not be found in the metadata and has not been updated."
            )

        if out is None:
            return m
        else:
            out.events.data_changed.trigger(obj=out)
        return m

    rebin.__doc__ = BaseSignal.rebin.__doc__

    def set_elements(self, elements):
        """Erase all elements and set them.

        Parameters
        ----------
        elements : list of strings
            A list of chemical element symbols.

        See also
        --------
        add_elements, set_lines, add_lines

        Examples
        --------
        >>> s = exspy.data.EDS_SEM_TM002()
        >>> print(s.metadata.Sample.elements)
        >>> s.set_elements(['Al'])
        >>> print(s.metadata.Sample.elements)
        ['Al' 'C' 'Cu' 'Mn' 'Zr']
        ['Al']

        """
        # Erase previous elements and X-ray lines
        if "Sample.elements" in self.metadata:
            del self.metadata.Sample.elements
        self.add_elements(elements)

    def add_elements(self, elements):
        """Add elements and the corresponding X-ray lines.

        The list of elements is stored in `metadata.Sample.elements`

        Parameters
        ----------
        elements : list of strings
            The symbol of the elements.

        Examples
        --------
        >>> s = exspy.data.EDS_SEM_TM002()
        >>> print(s.metadata.Sample.elements)
        >>> s.add_elements(['Ar'])
        >>> print(s.metadata.Sample.elements)
        ['Al' 'C' 'Cu' 'Mn' 'Zr']
        ['Al', 'Ar', 'C', 'Cu', 'Mn', 'Zr']

        See also
        --------
        set_elements, add_lines, set_lines

        """
        if not isiterable(elements) or isinstance(elements, str):
            raise ValueError(
                "Input must be in the form of a list. For example, "
                "if `s` is the variable containing this EDS spectrum:\n "
                ">>> s.add_elements(('C',))\n"
                "See the docstring for more information."
            )
        if "Sample.elements" in self.metadata:
            elements_ = set(self.metadata.Sample.elements)
        else:
            elements_ = set()
        for element in elements:
            if element in elements_db:
                elements_.add(element)
            else:
                raise ValueError(f"{element} is not a valid chemical element symbol.")
        self.metadata.set_item("Sample.elements", sorted(list(elements_)))

    def _get_xray_lines(self, xray_lines=None, only_one=None, only_lines=("a",)):
        if xray_lines is None:
            if "Sample.xray_lines" in self.metadata:
                xray_lines = self.metadata.Sample.xray_lines
            elif "Sample.elements" in self.metadata:
                xray_lines = self._get_lines_from_elements(
                    self.metadata.Sample.elements,
                    only_one=only_one,
                    only_lines=only_lines,
                )
            else:
                raise ValueError("Not X-ray line, set them with `add_elements`.")
        return xray_lines

    def set_lines(self, lines, only_one=True, only_lines=("a",)):
        """Erase all Xrays lines and set them.

        See add_lines for details.

        Parameters
        ----------
        lines : list of strings
            A list of valid element X-ray lines to add e.g. Fe_Kb.
            Additionally, if `metadata.Sample.elements` is
            defined, add the lines of those elements that where not
            given in this list.
        only_one: bool
            If False, add all the lines of each element in
            `metadata.Sample.elements` that has not line
            defined in lines. If True (default),
            only add the line at the highest energy
            above an overvoltage of 2 (< beam energy / 2).
        only_lines : {None, list of strings}
            If not None, only the given lines will be added.

        Examples
        --------
        >>> s = exspy.data.EDS_SEM_TM002()
        >>> s.add_lines()
        >>> print(s.metadata.Sample.xray_lines)
        >>> s.set_lines(['Cu_Ka'])
        >>> print(s.metadata.Sample.xray_lines)
        ['Al_Ka', 'C_Ka', 'Cu_La', 'Mn_La', 'Zr_La']
        ['Al_Ka', 'C_Ka', 'Cu_Ka', 'Mn_La', 'Zr_La']

        See also
        --------
        add_lines, add_elements, set_elements

        """
        only_lines = utils_eds._parse_only_lines(only_lines)
        if "Sample.xray_lines" in self.metadata:
            del self.metadata.Sample.xray_lines
        self.add_lines(lines=lines, only_one=only_one, only_lines=only_lines)

    def add_lines(self, lines=(), only_one=True, only_lines=("a",)):
        """Add X-rays lines to the internal list.

        Although most functions do not require an internal list of
        X-ray lines because they can be calculated from the internal
        list of elements, ocassionally it might be useful to customize the
        X-ray lines to be use by all functions by default using this method.
        The list of X-ray lines is stored in
        `metadata.Sample.xray_lines`

        Parameters
        ----------
        lines : list of strings
            A list of valid element X-ray lines to add e.g. Fe_Kb.
            Additionally, if `metadata.Sample.elements` is
            defined, add the lines of those elements that where not
            given in this list. If the list is empty (default), and
            `metadata.Sample.elements` is
            defined, add the lines of all those elements.
        only_one: bool
            If False, add all the lines of each element in
            `metadata.Sample.elements` that has not line
            defined in lines. If True (default),
            only add the line at the highest energy
            above an overvoltage of 2 (< beam energy / 2).
        only_lines : {None, list of strings}
            If not None, only the given lines will be added.

        Examples
        --------
        >>> s = exspy.data.EDS_SEM_TM002()
        >>> s.add_lines()
        >>> print(s.metadata.Sample.xray_lines)
        ['Al_Ka', 'C_Ka', 'Cu_La', 'Mn_La', 'Zr_La']

        >>> s = exspy.data.EDS_SEM_TM002()
        >>> s.set_microscope_parameters(beam_energy=30)
        >>> s.add_lines()
        >>> print(s.metadata.Sample.xray_lines)
        ['Al_Ka', 'C_Ka', 'Cu_Ka', 'Mn_Ka', 'Zr_La']

        >>> s = exspy.data.EDS_SEM_TM002()
        >>> s.add_lines()
        >>> print(s.metadata.Sample.xray_lines)
        >>> s.add_lines(['Cu_Ka'])
        >>> print(s.metadata.Sample.xray_lines)
        ['Al_Ka', 'C_Ka', 'Cu_La', 'Mn_La', 'Zr_La']
        ['Al_Ka', 'C_Ka', 'Cu_Ka', 'Cu_La', 'Mn_La', 'Zr_La']

        See also
        --------
        set_lines, add_elements, set_elements

        """
        only_lines = utils_eds._parse_only_lines(only_lines)
        if "Sample.xray_lines" in self.metadata:
            xray_lines = set(self.metadata.Sample.xray_lines)
        else:
            xray_lines = set()
        # Define the elements which Xray lines has been customized
        # So that we don't attempt to add new lines automatically
        elements = set()
        for line in xray_lines:
            elements.add(line.split("_")[0])
        for line in lines:
            try:
                element, subshell = line.split("_")
            except ValueError:
                raise ValueError(
                    "Invalid line symbol. "
                    "Please provide a valid line symbol e.g. Fe_Ka"
                )
            if element in elements_db:
                elements.add(element)
                if subshell in elements_db[element]["Atomic_properties"]["Xray_lines"]:
                    lines_len = len(xray_lines)
                    xray_lines.add(line)
                    if lines_len != len(xray_lines):
                        _logger.info(f"{line} line added,")
                    else:
                        _logger.info(f"{line} line already in.")
                else:
                    raise ValueError(f"{line} is not a valid line of {element}.")
            else:
                raise ValueError(f"{element} is not a valid symbol of an element.")
        xray_not_here = self._get_xray_lines_in_spectral_range(xray_lines)[1]
        for xray in xray_not_here:
            warnings.warn(f"{xray} is not in the data energy range.", UserWarning)
        if "Sample.elements" in self.metadata:
            extra_elements = set(self.metadata.Sample.elements) - elements
            if extra_elements:
                new_lines = self._get_lines_from_elements(
                    extra_elements, only_one=only_one, only_lines=only_lines
                )
                if new_lines:
                    self.add_lines(list(new_lines) + list(lines))
        self.add_elements(elements)
        if not hasattr(self.metadata, "Sample"):
            self.metadata.add_node("Sample")
        if "Sample.xray_lines" in self.metadata:
            xray_lines = xray_lines.union(self.metadata.Sample.xray_lines)
        self.metadata.Sample.xray_lines = sorted(list(xray_lines))

    def _get_lines_from_elements(self, elements, only_one=False, only_lines=("a",)):
        """Returns the X-ray lines of the given elements in spectral range
        of the data.

        Parameters
        ----------
        elements : list of strings
            A list containing the symbol of the chemical elements.
        only_one : bool
            If False, add all the lines of each element in the data spectral
            range. If True only add the line at the highest energy
            above an overvoltage of 2 (< beam energy / 2).
        only_lines : {None, list of strings}
            If not None, only the given lines will be returned.

        Returns
        -------
        list of X-ray lines alphabetically sorted

        """

        only_lines = utils_eds._parse_only_lines(only_lines)
        try:
            beam_energy = self._get_beam_energy()
        except BaseException:
            # Fall back to the high_value of the energy axis
            beam_energy = self.axes_manager.signal_axes[0].high_value
        lines = []
        elements = [el if isinstance(el, str) else el.decode() for el in elements]
        for element in elements:
            # Possible line (existing and excited by electron)
            element_lines = []
            for subshell in list(
                elements_db[element]["Atomic_properties"]["Xray_lines"].keys()
            ):
                if only_lines and subshell not in only_lines:
                    continue
                element_lines.append(element + "_" + subshell)
            element_lines = self._get_xray_lines_in_spectral_range(element_lines)[0]
            if only_one and element_lines:
                # Choose the best line
                select_this = -1
                element_lines.sort()
                for i, line in enumerate(element_lines):
                    if self._get_line_energy(line) < beam_energy / 2:
                        select_this = i
                        break
                element_lines = [
                    element_lines[select_this],
                ]

            if not element_lines:
                _logger.info(
                    f"There is no X-ray line for element {element} "
                    "in the data spectral range"
                )
            else:
                lines.extend(element_lines)
        lines.sort()
        return lines

    def _parse_xray_lines(self, xray_lines, only_one, only_lines):
        only_lines = utils_eds._parse_only_lines(only_lines)
        xray_lines = self._get_xray_lines(
            xray_lines, only_one=only_one, only_lines=only_lines
        )
        xray_lines, xray_not_here = self._get_xray_lines_in_spectral_range(xray_lines)
        for xray in xray_not_here:
            warnings.warn(
                f"{xray} is not in the data energy range. "
                "You can remove it with: "
                f"`s.metadata.Sample.xray_lines.remove('{xray}')`"
            )
        return xray_lines

    def get_lines_intensity(
        self,
        xray_lines=None,
        integration_windows=2.0,
        background_windows=None,
        plot_result=False,
        only_one=True,
        only_lines=("a",),
        **kwargs,
    ):
        """Return the intensity map of selected Xray lines.

        The intensities, the number of X-ray counts, are computed by
        suming the spectrum over the
        different X-ray lines. The sum window width
        is calculated from the energy resolution of the detector
        as defined in 'energy_resolution_MnKa' of the metadata.
        Backgrounds average in provided windows can be subtracted from the
        intensities.

        Parameters
        ----------
        xray_lines: {None, Iterable* of strings}
            If None,
            if `metadata.Sample.elements.xray_lines` contains a
            list of lines use those.
            If `metadata.Sample.elements.xray_lines` is undefined
            or empty but `metadata.Sample.elements` is defined,
            use the same syntax as `add_line` to select a subset of lines
            for the operation.
            Alternatively, provide an iterable containing
            a list of valid X-ray lines symbols.
            * Note that while dictionaries and strings are iterable,
            their use is ambiguous and specifically not allowed.
        integration_windows: Float or array
            If float, the width of the integration windows is the
            'integration_windows_width' times the calculated FWHM of the line.
            Else provide an array for which each row corresponds to a X-ray
            line. Each row contains the left and right value of the window.
        background_windows: None or 2D array of float
            If None, no background subtraction. Else, the backgrounds average
            in the windows are subtracted from the return intensities.
            'background_windows' provides the position of the windows in
            energy. Each line corresponds to a X-ray line. In a line, the two
            first values correspond to the limits of the left window and the
            two last values correspond to the limits of the right window.
        plot_result : bool
            If True, plot the calculated line intensities. If the current
            object is a single spectrum it prints the result instead.
        only_one : bool
            If False, use all the lines of each element in the data spectral
            range. If True use only the line at the highest energy
            above an overvoltage of 2 (< beam energy / 2).
        only_lines : {None, list of strings}
            If not None, use only the given lines.
        kwargs
            The extra keyword arguments for plotting. See
            `utils.plot.plot_signals`

        Returns
        -------
        intensities : list
            A list containing the intensities as BaseSignal subclasses.

        Examples
        --------
        >>> s = exspy.data.EDS_SEM_TM002()
        >>> s.get_lines_intensity(['Mn_Ka'], plot_result=True)
        Mn_La at 0.63316 keV : Intensity = 96700.00

        >>> s = exspy.data.EDS_SEM_TM002()
        >>> s.plot(['Mn_Ka'], integration_windows=2.1)
        >>> s.get_lines_intensity(['Mn_Ka'],
        >>>                       integration_windows=2.1, plot_result=True)
        Mn_Ka at 5.8987 keV : Intensity = 53597.00

        >>> s = exspy.data.EDS_SEM_TM002()
        >>> s.set_elements(['Mn'])
        >>> s.set_lines(['Mn_Ka'])
        >>> bw = s.estimate_background_windows()
        >>> s.plot(background_windows=bw)
        >>> s.get_lines_intensity(background_windows=bw, plot_result=True)
        Mn_Ka at 5.8987 keV : Intensity = 46716.00

        See also
        --------
        set_elements, add_elements, estimate_background_windows,
        plot

        """
        if xray_lines is not None and (
            not isinstance(xray_lines, Iterable) or isinstance(xray_lines, (str, dict))
        ):
            raise TypeError(
                "xray_lines must be a compatible iterable, but was "
                f"mistakenly provided as a {type(xray_lines)}."
            )

        xray_lines = self._parse_xray_lines(xray_lines, only_one, only_lines)
        if hasattr(integration_windows, "__iter__") is False:
            integration_windows = self.estimate_integration_windows(
                windows_width=integration_windows, xray_lines=xray_lines
            )
        intensities = []
        ax = self.axes_manager.signal_axes[0]
        # test Signal1D (0D problem)
        # signal_to_index = self.axes_manager.navigation_dimension - 2
        for i, (Xray_line, window) in enumerate(zip(xray_lines, integration_windows)):
            element, line = utils_eds._get_element_and_line(Xray_line)
            line_energy = self._get_line_energy(Xray_line)
            # Replace with `map` function for lazy large datasets
            img = self.isig[window[0] : window[1]].integrate1D(
                -1
            )  # integrate over window.
            if np.issubdtype(img.data.dtype, np.integer):
                # The operations below require a float dtype with the default
                # numpy casting rule ('same_kind')
                img.change_dtype("float")
            if background_windows is not None:
                bw = background_windows[i]
                # TODO: test to prevent slicing bug. To be reomved when fixed
                indexes = [float(ax.value2index(de)) for de in list(bw) + window]
                if indexes[0] == indexes[1]:
                    bck1 = self.isig[bw[0]]
                else:
                    bck1 = self.isig[bw[0] : bw[1]].integrate1D(-1)
                if indexes[2] == indexes[3]:
                    bck2 = self.isig[bw[2]]
                else:
                    bck2 = self.isig[bw[2] : bw[3]].integrate1D(-1)
                corr_factor = (indexes[5] - indexes[4]) / (
                    (indexes[1] - indexes[0]) + (indexes[3] - indexes[2])
                )
                img = img - (bck1 + bck2) * corr_factor
            img.metadata.General.title = (
                f"X-ray line intensity of {self.metadata.General.title}: "
                f"{Xray_line} at {line_energy:.2f} "
                f"{self.axes_manager.signal_axes[0].units}"
            )
            img = img.transpose(signal_axes=[])
            if plot_result and img.axes_manager.navigation_size == 1:
                if img._lazy:
                    img.compute()
                print(
                    f"{Xray_line} at {line_energy} {ax.units} : "
                    f"Intensity = {img.data[0]:.2f}"
                )
            img.metadata.set_item("Sample.elements", ([element]))
            img.metadata.set_item("Sample.xray_lines", ([Xray_line]))
            intensities.append(img)
        if plot_result and img.axes_manager.navigation_size != 1:
            utils.plot.plot_signals(intensities, **kwargs)
        return intensities

    def get_take_off_angle(self):
        """Calculate the take-off-angle (TOA).

        TOA is the angle with which the X-rays leave the surface towards
        the detector. Parameters are read in 'SEM.Stage.tilt_alpha',
        'Acquisition_instrument.SEM.Detector.EDS.azimuth_angle' and
        'SEM.Detector.EDS.elevation_angle' and 'SEM.Stage.tilt_beta in
        'metadata'.

        Returns
        -------
        take_off_angle: float
            in Degree

        Examples
        --------
        >>> s = exspy.data.EDS_SEM_TM002()
        >>> s.get_take_off_angle()
        37.0
        >>> s.set_microscope_parameters(tilt_stage=20.)
        >>> s.get_take_off_angle()
        57.0

        See also
        --------
        hs.eds.take_off_angle
        """
        if self.metadata.Signal.signal_type == "EDS_SEM":
            mp = self.metadata.Acquisition_instrument.SEM
        elif self.metadata.Signal.signal_type == "EDS_TEM":
            mp = self.metadata.Acquisition_instrument.TEM

        tilt_stage = mp.get_item("Stage.tilt_alpha", None)
        azimuth_angle = mp.get_item("Detector.EDS.azimuth_angle", None)
        elevation_angle = mp.get_item("Detector.EDS.elevation_angle", None)
        beta_tilt = mp.get_item("Stage.tilt_beta", 0.0)

        return utils_eds.take_off_angle(
            tilt_stage, azimuth_angle, elevation_angle, beta_tilt
        )

    def estimate_integration_windows(self, windows_width=2.0, xray_lines=None):
        """
        Estimate a window of integration for each X-ray line.

        Parameters
        ----------
        windows_width: float
            The width of the integration windows is the 'windows_width' times
            the calculated FWHM of the line.
        xray_lines: None or list of string
            If None, use 'metadata.Sample.elements.xray_lines'. Else,
            provide an iterable containing a list of valid X-ray lines
            symbols.

        Return
        ------
        integration_windows: 2D array of float
            The positions of the windows in energy. Each row corresponds to a
            X-ray line. Each row contains the left and right value of the
            window.

        Examples
        --------
        >>> s = exspy.data.EDS_SEM_TM002()
        >>> s.add_lines()
        >>> iw = s.estimate_integration_windows()
        >>> s.plot(integration_windows=iw)
        >>> s.get_lines_intensity(integration_windows=iw, plot_result=True)
        Fe_Ka at 6.4039 keV : Intensity = 3710.00
        Pt_La at 9.4421 keV : Intensity = 15872.00

        See also
        --------
        plot, get_lines_intensity
        """
        xray_lines = self._get_xray_lines(xray_lines)
        integration_windows = []
        for Xray_line in xray_lines:
            line_energy, line_FWHM = self._get_line_energy(Xray_line, FWHM_MnKa="auto")
            element, line = utils_eds._get_element_and_line(Xray_line)
            det = windows_width * line_FWHM / 2.0
            integration_windows.append([line_energy - det, line_energy + det])
        return integration_windows

    def estimate_background_windows(
        self, line_width=[2, 2], windows_width=1, xray_lines=None
    ):
        """
        Estimate two windows around each X-ray line containing only the
        background.

        Parameters
        ----------
        line_width: list of two floats
            The position of the two windows around the X-ray line is given by
            the `line_width` (left and right) times the calculated FWHM of the
            line.
        windows_width: float
            The width of the windows is is the `windows_width` times the
            calculated FWHM of the line.
        xray_lines: None or list of string
            If None, use `metadata.Sample.elements.xray_lines`. Else,
            provide an iterable containing a list of valid X-ray lines
            symbols.

        Return
        ------
        windows_position: 2D array of float
            The position of the windows in energy. Each line corresponds to a
            X-ray line. In a line, the two first values correspond to the
            limits of the left window and the two last values correspond to
            the limits of the right window.

        Examples
        --------
        >>> s = exspy.data.EDS_SEM_TM002()
        >>> s.add_lines()
        >>> bw = s.estimate_background_windows(line_width=[5.0, 2.0])
        >>> s.plot(background_windows=bw)
        >>> s.get_lines_intensity(background_windows=bw, plot_result=True)
        Fe_Ka at 6.4039 keV : Intensity = 2754.00
        Pt_La at 9.4421 keV : Intensity = 15090.00

        See also
        --------
        plot, get_lines_intensity
        """
        xray_lines = self._get_xray_lines(xray_lines)
        windows_position = []
        for xray_line in xray_lines:
            line_energy, line_FWHM = self._get_line_energy(xray_line, FWHM_MnKa="auto")
            tmp = [
                line_energy - line_FWHM * line_width[0] - line_FWHM * windows_width,
                line_energy - line_FWHM * line_width[0],
                line_energy + line_FWHM * line_width[1],
                line_energy + line_FWHM * line_width[1] + line_FWHM * windows_width,
            ]
            windows_position.append(tmp)
        windows_position = np.array(windows_position)
        # merge ovelapping windows
        index = windows_position.argsort(axis=0)[:, 0]
        for i in range(len(index) - 1):
            ia, ib = index[i], index[i + 1]
            if windows_position[ia, 2] > windows_position[ib, 0]:
                interv = np.append(windows_position[ia, :2], windows_position[ib, 2:])
                windows_position[ia] = interv
                windows_position[ib] = interv
        return windows_position

    def plot(
        self,
        xray_lines=False,
        only_lines=("a", "b"),
        only_one=False,
        background_windows=None,
        integration_windows=None,
        navigator="auto",
        plot_markers=True,
        autoscale="v",
        norm="auto",
        axes_manager=None,
        navigator_kwds={},
        **kwargs,
    ):
        """Plot the EDS spectrum. The following markers can be added

        - The position of the X-ray lines and their names.
        - The background windows associated with each X-ray lines. A black line
          links the left and right window with the average value in each window.

        Parameters
        ----------
        xray_lines: {False, True, 'from_elements', list of string}
            If not False, indicate the position and the name of the X-ray
            lines.
            If True, if `metadata.Sample.elements.xray_lines` contains a
            list of lines use those. If `metadata.Sample.elements.xray_lines`
            is undefined or empty or if xray_lines equals 'from_elements' and
            `metadata.Sample.elements` is defined, use the same syntax as
            `add_line` to select a subset of lines for the operation.
            Alternatively, provide an iterable containing a list of valid X-ray
            lines symbols.
        only_lines : None or list of strings
            If not None, use only the given lines (eg. ('a','Kb')).
            If None, use all lines.
        only_one : bool
            If False, use all the lines of each element in the data spectral
            range. If True use only the line at the highest energy
            above an overvoltage of 2 (< beam energy / 2).
        background_windows: None or 2D array of float
            If not None, add markers at the position of the windows in energy.
            Each line corresponds to a X-ray lines. In a line, the two first
            value corresponds to the limit of the left window and the two
            last values corresponds to the limit of the right window.
        integration_windows: None or 'auto' or float or 2D array of float
            If not None, add markers at the position of the integration
            windows.
            If 'auto' (or float), the width of the integration windows is 2.0
            (or float) times the calculated FWHM of the line. see
            'estimate_integration_windows'.
            Else provide an array for which each row corresponds to a X-ray
            line. Each row contains the left and right value of the window.
        %s
        %s

        Examples
        --------
        >>> s = exspy.data.EDS_SEM_TM002()
        >>> s.plot()

        >>> s = exspy.data.EDS_SEM_TM002()
        >>> s.plot(True)

        >>> s = exspy.data.EDS_SEM_TM002()
        >>> s.add_lines()
        >>> bw = s.estimate_background_windows()
        >>> s.plot(background_windows=bw)

        >>> s = exspy.data.EDS_SEM_TM002()
        >>> s.plot(['Mn_Ka'], integration_windows='auto')

        >>> s = exspy.data.EDS_SEM_TM002()
        >>> s.add_lines()
        >>> bw = s.estimate_background_windows()
        >>> s.plot(background_windows=bw, integration_windows=2.1)

        See also
        --------
        set_elements, add_elements, estimate_integration_windows,
        get_lines_intensity, estimate_background_windows
        """
        super().plot(
            navigator=navigator,
            plot_markers=plot_markers,
            autoscale=autoscale,
            norm=norm,
            axes_manager=axes_manager,
            navigator_kwds=navigator_kwds,
            **kwargs,
        )
        self._plot_xray_lines(
            xray_lines,
            only_lines,
            only_one,
            background_windows,
            integration_windows,
            render_figure=False,
        )
        self._render_figure(plot=["signal_plot"])

    plot.__doc__ %= (BASE_PLOT_DOCSTRING_PARAMETERS, PLOT1D_DOCSTRING)

    def _plot_xray_lines(
        self,
        xray_lines=False,
        only_lines=("a", "b"),
        only_one=False,
        background_windows=None,
        integration_windows=None,
        render_figure=True,
    ):
        if (
            xray_lines is not False
            or background_windows is not None
            or integration_windows is not None
        ):
            if xray_lines is False:
                xray_lines = True
            only_lines = utils_eds._parse_only_lines(only_lines)
            if xray_lines is True or xray_lines == "from_elements":
                if (
                    "Sample.xray_lines" in self.metadata
                    and xray_lines != "from_elements"
                ):
                    xray_lines = self.metadata.Sample.xray_lines
                elif "Sample.elements" in self.metadata:
                    xray_lines = self._get_lines_from_elements(
                        self.metadata.Sample.elements,
                        only_one=only_one,
                        only_lines=only_lines,
                    )
                else:
                    _logger.warning("No elements defined, set them with `add_elements`")
                    # No X-rays lines, nothing to do then
                    return

            xray_lines, xray_not_here = self._get_xray_lines_in_spectral_range(
                xray_lines
            )
            for xray in xray_not_here:
                _logger.warning(f"{xray} is not in the data energy range.")

            xray_lines = np.unique(xray_lines)

            self.add_xray_lines_markers(xray_lines, render_figure=False)
            if background_windows is not None:
                self._add_background_windows_markers(
                    background_windows, render_figure=False
                )
            if integration_windows is not None:
                if integration_windows == "auto":
                    integration_windows = 2.0
                if hasattr(integration_windows, "__iter__") is False:
                    integration_windows = self.estimate_integration_windows(
                        windows_width=integration_windows, xray_lines=xray_lines
                    )
                self._add_vertical_lines_groups(
                    integration_windows, linestyle="--", render_figure=False
                )
        # Render figure only at the end
        if render_figure:
            self._render_figure(plot=["signal_plot"])

    def _add_vertical_lines_groups(self, position, render_figure=True, **kwargs):
        """
        Add vertical markers for each group that shares the color.

        Parameters
        ----------
        position: 2D array of float
            The position on the signal axis. Each row corresponds to a
            group.
        kwargs
            keywords argument for :py:class:`~.api.plot.markers.VerticalLine`
        """
        colors = itertools.cycle(
            np.sort(plt.rcParams["axes.prop_cycle"].by_key()["color"])
        )

        for x, color in zip(position, colors):
            line = VerticalLines(offsets=x, color=color, **kwargs)
            self.add_marker(line, render_figure=False)
        if render_figure:
            self._render_figure(plot=["signal_plot"])

    def add_xray_lines_markers(self, xray_lines, render_figure=True):
        """
        Add marker on a spec.plot() with the name of the selected X-ray
        lines

        Parameters
        ----------
        xray_lines: list of string
            A valid list of X-ray lines
        """
        if self._plot is None or not self._plot.is_active:
            raise RuntimeError("The signal needs to be plotted.")
        norm = self._plot.signal_plot.ax_lines[0].norm
        minimum_intensity = self.data[self.data > 0].min() if norm == "log" else 0
        line_names = []
        segments = np.empty((len(xray_lines), 2, 2))
        offsets = np.empty((len(xray_lines), 2))
        # might want to set the intensity based on the alpha line intensity
        for i, xray_line in enumerate(xray_lines):
            element, line = utils_eds._get_element_and_line(xray_line)
            relative_factor = elements_db[element]["Atomic_properties"]["Xray_lines"][
                line
            ]["weight"]
            eng = self._get_line_energy(f"{element}_{line}")
            segments[i] = [[eng, 0], [eng, 1]]
            offsets[i] = [eng, 1]
            line_names.append(
                r"$\mathrm{%s}_{\mathrm{%s}}$"
                % utils_eds._get_element_and_line(xray_line)
            )

        line_markers = Lines(
            segments=segments,
            transform="relative",
            color="black",
        )
        text_markers = Texts(
            offsets=offsets,
            texts=line_names,
            offset_transform="relative",
            rotation=np.pi / 2,
            horizontalalignment="left",
            verticalalignment="bottom",
            facecolor="black",
            shift=0.005,
        )

        self.add_marker(line_markers, render_figure=False)
        self.add_marker(text_markers, render_figure=False)

        # Connect events to remove the markers when the line is closed
        line_markers.events.closed.connect(self._xray_marker_closed)
        text_markers.events.closed.connect(self._xray_marker_closed)
        self._xray_markers["lines"] = line_markers
        self._xray_markers["texts"] = text_markers
        self._xray_markers["names"] = xray_lines

        if render_figure:
            self._render_figure(plot=["signal_plot"])

    def _xray_marker_closed(self, obj):
        self._xray_markers = {}

    def remove_xray_lines_markers(self, xray_lines, render_figure=True):
        """
        Remove marker previously added on a spec.plot() with the name of the
        selected X-ray lines

        Parameters
        ----------
        xray_lines: list of string
            A valid list of X-ray lines to remove
        render_figure: bool
            If True, render the figure after removing the markers
        """
        ind = np.where(np.isin(self._xray_markers["names"], xray_lines))
        self._xray_markers["lines"].remove_items(ind)
        self._xray_markers["texts"].remove_items(ind)
        self._xray_markers["names"] = np.delete(self._xray_markers["names"], ind)
        if render_figure:
            self._render_figure(plot=["signal_plot"])

    def _add_background_windows_markers(self, windows_position, render_figure=True):
        """
        Plot the background windows associated with each X-ray lines.

        For X-ray lines, a black line links the left and right window with the
        average value in each window.

        Parameters
        ----------
        windows_position: 2D array of float
            The position of the windows in energy. Each line corresponds to a
            X-ray lines. In a line, the two first value corresponds to the
            limit of the left window and the two last values corresponds to the
            limit of the right window.

        See also
        --------
        estimate_background_windows, get_lines_intensity
        """
        self._add_vertical_lines_groups(windows_position)
        ax = self.axes_manager.signal_axes[0]
        segments = []
        for bw in windows_position:
            # TODO: test to prevent slicing bug. To be removed when fixed
            if ax.value2index(bw[0]) == ax.value2index(bw[1]):
                y1 = self.isig[bw[0]].data
            else:
                y1 = self.isig[bw[0] : bw[1]].mean(-1).data
            if ax.value2index(bw[2]) == ax.value2index(bw[3]):
                y2 = self.isig[bw[2]].data
            else:
                y2 = self.isig[bw[2] : bw[3]].mean(-1).data
            x1 = (bw[0] + bw[1]) / 2.0
            x2 = (bw[2] + bw[3]) / 2.0
            segments.append([[x1, y1[0]], [x2, y2[0]]])
        segments = np.array(segments)
        lines = Lines(segments=segments, color="black")
        self.add_marker(lines, render_figure=False)
        if render_figure:
            self._render_figure(plot=["signal_plot"])


class LazyEDSSpectrum(EDSSpectrum, LazySignal1D):
    """Lazy general signal class for EDS spectra."""

    __doc__ += LAZYSIGNAL_DOC.replace("__BASECLASS__", "EDSSpectrum")



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))

import hyperspy.api as hs
import logging
import numpydoc
from packaging.version import Version


# Set logging level to `ERROR` to avoid exspy warning in documentation
hs.set_log_level("ERROR")


# -- Project information -----------------------------------------------------

project = "exSpy"
copyright = "2023, exSpy Developers"
author = "exSpy Developers"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    # numpydoc is necessary to parse the docstring using sphinx
    # otherwise the nitpicky option will raise many warnings
    # "numpydoc",
    "sphinx_design",
    "sphinx_favicon",
    "sphinx_copybutton",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.githubpages",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx_gallery.gen_gallery",
    "sphinxcontrib.towncrier",
]

linkcheck_ignore = [
    "https://doi.org/10.1021/acs.nanolett.5b00449",  # 403 Client Error: Forbidden for url
    "https://onlinelibrary.wiley.com/doi/10.1111/j.1365-2818.2006.01549.x",  # 403 Client Error: Forbidden for url
    # Remove when setup
    "https://github.com/hyperspy/exspy-demos",
    "https://www.anaconda.com/blog/understanding-conda-and-pip",  # Transcient?
    # Remove once it is merged and the links are working
    "https://exspy.readthedocs.io",
    "https://github.com/hyperspy/exspy/blob/main/releasing_guide.md",
]

intersphinx_mapping = {
    "dask": ("https://docs.dask.org/en/latest", None),
    "hyperspy": ("https://hyperspy.org/hyperspy-doc/dev", None),
    "kikuchipy": ("https://kikuchipy.org/en/latest/", None),
    "matplotlib": ("https://matplotlib.org/stable", None),
    "numpy": ("https://numpy.org/doc/stable", None),
    "python": ("https://docs.python.org/3", None),
    "rsciio": ("https://hyperspy.org/rosettasciio/", None),
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "pydata_sphinx_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

html_theme_options = {
    "show_toc_level": 2,
    "github_url": "https://github.com/hyperspy/exspy",
    "icon_links": [
        {
            "name": "Gitter",
            "url": "https://gitter.im/hyperspy/hyperspy",
            "icon": "fab fa-gitter",
        },
    ],
    "logo": {
        "text": "exSpy",
    },
    "header_links_before_dropdown": 6,
    "navigation_with_keys": False,
}

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = "_static/hyperspy_logo.png"

# -- Options for sphinx_favicon extension -----------------------------------

favicons = [
    "hyperspy.ico",
]

# Check links to API when building documentation
nitpicky = False
# Remove when fixed in hyperspy
nitpick_ignore_regex = [(r"py:.*", r"hyperspy.api.*")]

# -- Options for numpydoc extension -----------------------------------

numpydoc_xref_param_type = True
numpydoc_xref_ignore = {"type", "optional", "default", "of"}

if Version(numpydoc.__version__) >= Version("1.6.0rc0"):
    numpydoc_validation_checks = {"all", "ES01", "EX01", "GL02", "GL03", "SA01", "SS06"}

autoclass_content = "both"

autodoc_default_options = {
    "show-inheritance": True,
}
toc_object_entries_show_parents = "hide"

# -- Sphinx-Gallery---------------

# https://sphinx-gallery.github.io
sphinx_gallery_conf = {
    "examples_dirs": "../examples",  # path to your example scripts
    "gallery_dirs": "auto_examples",  # path to where to save gallery generated output
    "filename_pattern": ".py",  # pattern to define which will be executed
    "ignore_pattern": "_sgskip.py",  # pattern to define which will not be executed
}

# -- Sphinx-copybutton -----------

copybutton_prompt_text = r">>> |\.\.\. "
copybutton_prompt_is_regexp = True

# -- Options for towncrier_draft extension -----------------------------------

# Options: draft/sphinx-version/sphinx-release
towncrier_draft_autoversion_mode = "draft"
towncrier_draft_include_empty = False
towncrier_draft_working_directory = ".."



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The exSpy developers
#
# This file is part of exSpy.
#
# exSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

import numpy as np

from exspy.components import EELSArctan


def test_function():
    g = EELSArctan()
    g.A.value = 10
    g.k.value = 2
    g.x0.value = 1
    np.testing.assert_allclose(g.function(0), 4.63647609)
    np.testing.assert_allclose(g.function(1), 10 * np.pi / 2)
    np.testing.assert_allclose(g.function(1e4), 10 * np.pi, 1e-4)
